"""
Tartarus Chroma Effects
"""
from openrazer_daemon.dbus_services import endpoint

@endpoint('razer.device.lighting.profile_led', 'getRedLED', out_sig='b')
def tartarus_get_profile_led_red(self):
    """
    Get red profile LED state

    :return: Red profile LED state
    :rtype: bool
    """
    self.logger.debug("DBus call tartarus_profile_led_red")

    driver_path = self.get_driver_path('profile_led_red')

    with open(driver_path, 'r') as driver_file:
        return driver_file.read().strip() == '1'


@endpoint('razer.device.lighting.profile_led', 'setRedLED', in_sig='b')
def tartarus_set_profile_led_red(self, enable):
    """
    Set red profile LED state

    :param enable: Status of red profile LED
    :type enable: bool
    """
    self.logger.debug("DBus call tartarus_set_profile_led_red")

    driver_path = self.get_driver_path('profile_led_red')

    with open(driver_path, 'w') as driver_file:
        if enable:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.lighting.profile_led', 'getGreenLED', out_sig='b')
def tartarus_get_profile_led_green(self):
    """
    Get green profile LED state

    :return: Green profile LED state
    :rtype: bool
    """
    self.logger.debug("DBus call tartarus_get_profile_led_green")

    driver_path = self.get_driver_path('profile_led_green')

    with open(driver_path, 'r') as driver_file:
        return driver_file.read().strip() == '1'


@endpoint('razer.device.lighting.profile_led', 'setGreenLED', in_sig='b')
def tartarus_set_profile_led_green(self, enable):
    """
    Set green profile LED state

    :param enable: Status of green profile LED
    :type enable: bool
    """
    self.logger.debug("DBus call tartarus_set_profile_led_green")

    driver_path = self.get_driver_path('profile_led_green')

    with open(driver_path, 'w') as driver_file:
        if enable:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.lighting.profile_led', 'getBlueLED', out_sig='b')
def tartarus_get_profile_led_blue(self):
    """
    Get blue profile LED state

    :return: Blue profile LED state
    :rtype: bool
    """
    self.logger.debug("DBus call tartarus_get_profile_led_blue")

    driver_path = self.get_driver_path('profile_led_blue')

    with open(driver_path, 'r') as driver_file:
        return driver_file.read().strip() == '1'


@endpoint('razer.device.lighting.profile_led', 'setBlueLED', in_sig='b')
def tartarus_set_profile_led_blue(self, enable):
    """
    Set blue profile LED state

    :param enable: Status of blue profile LED
    :type enable: bool
    """
    self.logger.debug("DBus call tartarus_set_profile_led_blue")

    driver_path = self.get_driver_path('profile_led_blue')

    with open(driver_path, 'w') as driver_file:
        if enable:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.macro', 'getModeModifier', out_sig='b')
def tartarus_get_mode_modifier(self):
    """
    Get if the mode key is a modifier

    :return: State
    :rtype: bool
    """
    self.logger.debug("DBus call tartarus_get_mode_modifier")

    return self.key_manager.mode_modifier


@endpoint('razer.device.macro', 'setModeModifier', in_sig='b')
def tartarus_set_mode_modifier(self, modifier):
    """
    Set if the mode key is a modifier

    :param modifier: State
    :type modifier: bool
    """
    self.logger.debug("DBus call tartarus_set_mode_modifier")

    self.key_manager.mode_modifier = modifier



import sys
import time
import typing

from openrazer.client.devices.keyboard import RazerKeyboard

import openrazer.client

COLS = 22
ROWS = 6


def find_keyboard() -> RazerKeyboard:
    device_manager = openrazer.client.DeviceManager()

    for device in device_manager.devices:
        if device.type == 'keyboard':
            keyboard = device
            break
    else:
        print("Could not find suitable keyboard", file=sys.stderr)
        sys.exit(1)

    if not keyboard.has('lighting_led_matrix'):
        print("Keyboard doesnt have LED matrix", file=sys.stderr)
        sys.exit(1)

    return keyboard


def index_to_coord(index: int) -> typing.Tuple[int, int]:
    index = index % (ROWS * COLS)
    col = index % COLS
    row = index // COLS
    return row, col




def int_to_color(v: int) -> typing.Tuple[int, int, int]:
    r = (v & 0xff0000) >> 16
    g = (v & 0x00ff00) >> 8
    b = (v & 0xff)
    return r, g, b


class FpsLimiter(object):
    def __init__(self, max_fps=60):
        self.frame_start = time.time()
        self.wait = 1
        self.set_limit(max_fps)

    def start_frame(self):
        self.frame_start = time.time()

    def wait_for_next_frame(self):
        delay = self.wait - (time.time() - self.frame_start)
        if delay > 0:
            time.sleep(delay)

    def set_limit(self, max_fps):
        self.wait = 1 / max_fps

    def current_fps(self):
        return 1 / (time.time() - self.frame_start)



import dbus as _dbus

from openrazer.client.devices import RazerDevice as __RazerDevice
from openrazer.client.macro import RazerMacro as _RazerMacro
from openrazer.client import constants as _c


class RazerMouse(__RazerDevice):
    _MACRO_CLASS = _RazerMacro

    def __init__(self, serial, vid_pid=None, daemon_dbus=None):
        super(RazerMouse, self).__init__(serial, vid_pid=vid_pid, daemon_dbus=daemon_dbus)

        # Capabilities
        self._capabilities['poll_rate'] = self._has_feature('razer.device.misc', ('getPollRate', 'setPollRate'))
        self._capabilities['dpi'] = self._has_feature('razer.device.dpi', ('getDPI', 'setDPI'))

        if self.has('dpi'):
            self._dbus_interfaces['dpi'] = _dbus.Interface(self._dbus, "razer.device.dpi")

    @property
    def max_dpi(self) -> int:
        """
        Gets max DPI

        :return: Max DPI, if device does not have DPI it'll return None
        :rtype: int or None
        """
        if self.has('dpi'):
            return int(self._dbus_interfaces['dpi'].maxDPI())
        else:
            return None

    @property
    def dpi(self) -> tuple:
        """
        Get mouse DPI

        Will return a tuple
        :return: DPI (500, 500)
        :rtype: tuple

        :raises NotImplementedError: If function is not supported
        """
        if self.has('dpi'):
            dpi_x, dpi_y = self._dbus_interfaces['dpi'].getDPI()
            # Converting to integers to remove the dbus types
            return int(dpi_x), int(dpi_y)
        else:
            raise NotImplementedError()

    @dpi.setter
    def dpi(self, value:tuple):
        """
        Set mouse dpi

        Daemon does type validation but cant be too careful
        :param value: DPI X, Y tuple
        :type value: tuple

        :raises ValueError: If the tuple isnt long enough or contains invalid crap
        :raises NotImplementedError: If function is not supported
        """
        if self.has('dpi'):
            if len(value) != 2:
                raise ValueError("DPI tuple is not of length 2. Length: {0}".format(len(value)))
            dpi_x, dpi_y = value

            if not isinstance(dpi_x, int) or not isinstance(dpi_y, int):
                raise ValueError("DPI X or Y is not an integer, X:{0} Y:{1}".format(type(dpi_x), type(dpi_y)))

            if dpi_x < 0 or dpi_x > 16000: # TODO add in max dpi option
                raise ValueError("DPI X either too small or too large, X:{0}".format(dpi_x))
            if dpi_y < 0 or dpi_y > 16000:  # TODO add in max dpi option
                raise ValueError("DPI Y either too small or too large, Y:{0}".format(dpi_y))

            self._dbus_interfaces['dpi'].setDPI(dpi_x, dpi_y)
        else:
            raise NotImplementedError()

    @property
    def poll_rate(self) -> int:
        """
        Get poll rate from device

        :return: Poll rate
        :rtype: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('poll_rate'):
            return int(self._dbus_interfaces['device'].getPollRate())
        else:
            raise NotImplementedError()


    @poll_rate.setter
    def poll_rate(self, poll_rate:int):
        """
        Set poll rate of device

        :param poll_rate: Polling rate
        :type poll_rate: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('poll_rate'):
            if not isinstance(poll_rate, int):
                raise ValueError("Poll rate is not an integer: {0}".format(poll_rate))
            if poll_rate not in (_c.POLL_125HZ, _c.POLL_500HZ, _c.POLL_1000HZ):
                raise ValueError('Poll rate "{0}" is not one of {1}'.format(poll_rate, (_c.POLL_125HZ, _c.POLL_500HZ, _c.POLL_1000HZ)))

            self._dbus_interfaces['device'].setPollRate(poll_rate)

        else:
            raise NotImplementedError()



"""
BlackWidow Ultimate 2013 effects
"""
from openrazer_daemon.dbus_services import endpoint

@endpoint('razer.device.lighting.bw2013', 'getEffect', out_sig='y')
def bw_get_effect(self):
    """
    Get current effect

    :return: Brightness
    :rtype: int
    """
    self.logger.debug("DBus call bw_get_effect")

    driver_path = self.get_driver_path('matrix_effect_pulsate')

    with open(driver_path, 'r') as driver_file:
        brightness = int(driver_file.read().strip())
        return brightness


@endpoint('razer.device.lighting.bw2013', 'setPulsate')
def bw_set_pulsate(self):
    """
    Set pulsate mode
    """
    self.logger.debug("DBus call bw_set_pulsate")

    driver_path = self.get_driver_path('matrix_effect_pulsate')

    with open(driver_path, 'w') as driver_file:
        driver_file.write('1')

    # Notify others
    self.send_effect_event('setPulsate')


@endpoint('razer.device.lighting.bw2013', 'setStatic')
def bw_set_static(self):
    """
    Set static mode
    """
    self.logger.debug("DBus call bw_set_static")

    driver_path = self.get_driver_path('matrix_effect_static')

    with open(driver_path, 'w') as driver_file:
        driver_file.write('1')

    # Notify others
    self.send_effect_event('setStatic')






from __future__ import print_function

import optparse
import sys

from ore.audio.capture import DeviceFinder, graph_all_devices
from ore.registry.registry import get_registry


class Program(object):

    def __init__(self):
        self.parser = optparse.OptionParser(usage="""usage: %prog command [args...] [--audio-device=<device>]

Valid commands:
    list               List all effects
    run <effect>       Run an effect
    audiolist          List audio devices that *should* work
    audiotest          Test all audio devices
    audiotest <device> Test a single audio device

""".strip())
        self.parser.add_option(
            "-a",
            "--audio-device",
            dest="audio_device",
            default="default",
            help="name of the audio device to pair with, if used.")

    def _parse_and_execute(self, args):
        (opts, args) = self.parser.parse_args(args)

        if len(args) < 1:
            self._error("command required.")

        cmd = args[0]
        if len(cmd) > 0 and cmd[0] != '_' and hasattr(self, cmd) and callable(getattr(self, cmd)):
            func = getattr(self, cmd)
            func(*args, **vars(opts))
        else:
            self._error("invalid command {}".format(cmd))

    def _error(self, msg):
        self.parser.error(msg)

    def audiolist(self, *args, **kwargs):
        import pyaudio
        p = pyaudio.PyAudio()

        device_finder = DeviceFinder(p)

        devices = list(device_finder.find_input_devices())
        if len(devices) == 0:
            print("Sorry, no devices found.")
            return

        print("** Audio Devices **")
        for dev in devices:
            print(dev["name"])

    def audiotest(self, *args, **kwargs):
        import pyaudio
        p = pyaudio.PyAudio()

        device_names = None
        if len(args) > 1:
            device_name = args[1]
            device_names = [device_name]

        graph_all_devices(p, device_names=device_names)

    def list(self, *args, **kwargs):
        reg = get_registry()
        for cli_name, effect in reg.items():

            effect = effect()

            details = {
                "cli": effect.get_cli_name(),
                "name": effect.get_name(),
                "author": effect.get_author(),
                "description": effect.get_description(),
                "type": effect.get_effect_type(),
            }
            if details["description"] is None:
                details["description"] = "No description."

            print("""{cli} :: {name} [{author}]
{description}
""".format(**details))

    def run(self, *args, **kwargs):
        if len(args) < 2:
            self._error("effect name required.")

        cli_name = args[1]
        reg = get_registry()
        if cli_name not in reg:
            self._error("effect {} not found".format(cli_name))

        effect_class = reg[cli_name]
        effect = effect_class(**kwargs)
        effect.start()


def main():
    prog = Program()
    prog._parse_and_execute(sys.argv[1:])


if __name__ == "__main__":
    main()



"""
Service Object for DBus
"""

# Disable some pylint stuff
# pylint: disable=no-member

import types
import dbus
import dbus.service


def copy_func(function_reference, name=None):
    """
    Copy function

    :param function_reference: Function
    :type function_reference: func

    :param name: Name of function
    :type name: str

    :return: Copy of function
    :rtype: func
    """
    if hasattr(function_reference, 'code'):
        return types.FunctionType(function_reference.code, function_reference.globals, name or function_reference.func_name, function_reference.defaults, function_reference.closure)
    else:
        return types.FunctionType(function_reference.__code__, function_reference.__globals__, name or function_reference.func_name, function_reference.__defaults__, function_reference.__closure__)


class DBusServiceFactory(object):
    """
    Factory object to create differnet service objects.

    Different service objects are useful, as the DBus service table is stored in the class which would be shared
    between all class instances.
    """
    service_number = 0

    @staticmethod
    def get_service(bus_name, object_path):
        """
        Get an instance of the service history

        :param bus_name: DBus bus name
        :type bus_name: str

        :param object_path: DBus Object name
        :type object_path: str

        :return: New object
        :rtype: DBusService
        """
        new_service = type("DBUSService{0:04}".format(DBusServiceFactory.service_number), (DBusService,), {})
        DBusServiceFactory.service_number += 1

        return new_service(bus_name, object_path)


class DBusService(dbus.service.Object):
    """
    DBus Service object

    Allows for dynamic method adding
    """
    BUS_TYPE = 'session'

    def __init__(self, bus_name, object_path):
        """
        Init the object

        :param bus_name: DBus bus name
        :type bus_name: str

        :param object_path: DBus Object name
        :type object_path: str
        """
        self.bus_name = bus_name
        self.object_path = object_path

        if DBusService.BUS_TYPE == 'session':
            bus_object = dbus.service.BusName(bus_name, bus=dbus.SessionBus())
        else:
            bus_object = dbus.service.BusName(bus_name, bus=dbus.SystemBus())

        super(DBusService, self).__init__(bus_object, object_path)

    def add_dbus_method(self, interface_name, function_name, function, in_signature=None, out_signature=None, byte_arrays=False):
        """
        Add method to DBus Object

        :param interface_name: DBus interface name
        :type interface_name: str

        :param function_name: DBus function name
        :type function_name: str

        :param function: Function reference
        :type function: object

        :param in_signature: DBus function signature
        :type in_signature: str

        :param out_signature: DBus function signature
        :type out_signature: str

        :param byte_arrays: Is byte array
        :type byte_arrays: bool
        """

        # Get class key for use in the DBus introspection table
        class_key = [key for key in self._dbus_class_table.keys() if key.endswith(self.__class__.__name__)][0]

        # Create a copy of the function so that if its used multiple times it wont affect other instances if the names changed
        function_deepcopy = copy_func(function, function_name)
        func = dbus.service.method(interface_name, in_signature=in_signature, out_signature=out_signature, byte_arrays=byte_arrays)(function_deepcopy)

        # Add method to DBus tables
        try:
            self._dbus_class_table[class_key][interface_name][function_name] = func
        except KeyError:
            self._dbus_class_table[class_key][interface_name] = {function_name: func}

        # Add method to class as DBus expects it to be there.
        setattr(self.__class__, function_name, func)

    def del_dbus_method(self, interface_name, function_name):
        """
        Remove method from DBus Object

        :param interface_name: DBus interface name
        :type interface_name: str

        :param function_name: DBus function name
        :type function_name: str
        """

        # Get class key for use in the DBus introspection table
        class_key = [key for key in self._dbus_class_table.keys() if key.endswith(self.__class__.__name__)][0]

        # Remove method from DBus tables
        # Remove method from class
        try:
            del self._dbus_class_table[class_key][interface_name][function_name]
            delattr(DBusService, function_name)

        except (KeyError, AttributeError):
            pass



from openrazer_daemon.dbus_services import endpoint


@endpoint('razer.device.lighting.backlight', 'getBacklightActive', out_sig='b')
def get_backlight_active(self):
    """
    Get if the backlight is lit up

    :return: Active
    :rtype: bool
    """
    self.logger.debug("DBus call get_backlight_active")

    driver_path = self.get_driver_path('backlight_led_state')

    with open(driver_path, 'r') as driver_file:
        active = int(driver_file.read().strip())
        return active == 1


@endpoint('razer.device.lighting.backlight', 'setBacklightActive', in_sig='b')
def set_backlight_active(self, active):
    """
    Get if the backlight is lit up

    :param active: Is active
    :type active: bool
    """
    self.logger.debug("DBus call set_backlight_active")

    driver_path = self.get_driver_path('backlight_led_state')

    with open(driver_path, 'w') as driver_file:
        if active:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.lighting.logo', 'getLogoActive', out_sig='b')
def get_logo_active(self):
    """
    Get if the logo is light up

    :return: Active
    :rtype: bool
    """
    self.logger.debug("DBus call get_logo_active")

    driver_path = self.get_driver_path('logo_led_state')

    with open(driver_path, 'r') as driver_file:
        active = int(driver_file.read().strip())
        return active == 1


@endpoint('razer.device.lighting.logo', 'setLogoActive', in_sig='b')
def set_logo_active(self, active):
    """
    Get if the logo is light up

    :param active: Is active
    :type active: bool
    """
    self.logger.debug("DBus call set_logo_active")

    driver_path = self.get_driver_path('logo_led_state')

    with open(driver_path, 'w') as driver_file:
        if active:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.lighting.logo', 'getLogoEffect', out_sig='y')
def get_logo_effect(self):
    """
    Get logo effect

    :return: Active
    :rtype: bool
    """
    self.logger.debug("DBus call get_logo_effect")

    driver_path = self.get_driver_path('logo_led_effect')

    with open(driver_path, 'r') as driver_file:
        effect = int(driver_file.read().strip())
        return effect


@endpoint('razer.device.lighting.logo', 'getLogoBrightness', out_sig='d')
def get_logo_brightness(self):
    """
    Get the device's brightness

    :return: Brightness
    :rtype: float
    """
    self.logger.debug("DBus call get_logo_brightness")

    driver_path = self.get_driver_path('logo_led_brightness')

    with open(driver_path, 'r') as driver_file:
        brightness = round(float(driver_file.read()) * (100.0/255.0), 2)

        return brightness


@endpoint('razer.device.lighting.logo', 'setLogoBrightness', in_sig='d')
def set_logo_brightness(self, brightness):
    """
    Set the device's brightness

    :param brightness: Brightness
    :type brightness: int
    """
    self.logger.debug("DBus call set_logo_brightness")

    driver_path = self.get_driver_path('logo_led_brightness')

    self.method_args['brightness'] = brightness

    brightness = int(round(brightness * (255.0/100.0)))
    if brightness > 255:
        brightness = 255
    elif brightness < 0:
        brightness = 0

    with open(driver_path, 'w') as driver_file:
        driver_file.write(str(brightness))

    # Notify others
    self.send_effect_event('setBrightness', brightness)


@endpoint('razer.device.lighting.logo', 'setLogoStatic', in_sig='yyy')
def set_logo_static(self, red, green, blue):
    """
    Set the device to static colour

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_static_effect")

    # Notify others
    self.send_effect_event('setStatic', red, green, blue)

    rgb_driver_path = self.get_driver_path('logo_led_rgb')
    effect_driver_path = self.get_driver_path('logo_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('0')


@endpoint('razer.device.lighting.logo', 'setLogoBlinking', in_sig='yyy')
def set_logo_blinking(self, red, green, blue):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_logo_blinking")

    # Notify others
    self.send_effect_event('setLogoBlinking', red, green, blue)

    rgb_driver_path = self.get_driver_path('logo_led_rgb')
    effect_driver_path = self.get_driver_path('logo_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('1')


@endpoint('razer.device.lighting.logo', 'setLogoPulsate', in_sig='yyy')
def set_logo_pulsate(self, red, green, blue):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_logo_pulsing")

    # Notify others
    self.send_effect_event('setPulsate', red, green, blue)

    rgb_driver_path = self.get_driver_path('logo_led_rgb')
    effect_driver_path = self.get_driver_path('logo_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('2')


@endpoint('razer.device.lighting.logo', 'setLogoSpectrum')
def set_logo_spectrum(self):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_logo_spectrum")

    # Notify others
    self.send_effect_event('setSpectrum')

    effect_driver_path = self.get_driver_path('logo_led_effect')

    with open(effect_driver_path, 'w') as effect_driver_file:
        effect_driver_file.write('4')


@endpoint('razer.device.lighting.scroll', 'getScrollActive', out_sig='b')
def get_scroll_active(self):
    """
    Get if the scroll is light up

    :return: Active
    :rtype: bool
    """
    self.logger.debug("DBus call get_scroll_active")

    driver_path = self.get_driver_path('scroll_led_state')

    with open(driver_path, 'r') as driver_file:
        active = int(driver_file.read().strip())
        return active == 1


@endpoint('razer.device.lighting.scroll', 'setScrollActive', in_sig='b')
def set_scroll_active(self, active):
    """
    Get if the scroll is light up

    :param active: Is active
    :type active: bool
    """
    self.logger.debug("DBus call set_scroll_active")

    driver_path = self.get_driver_path('scroll_led_state')

    with open(driver_path, 'w') as driver_file:
        if active:
            driver_file.write('1')
        else:
            driver_file.write('0')


@endpoint('razer.device.lighting.scroll', 'getScrollEffect', out_sig='y')
def get_scroll_effect(self):
    """
    Get scroll effect

    :return: Active
    :rtype: bool
    """
    self.logger.debug("DBus call get_scroll_effect")

    driver_path = self.get_driver_path('scroll_led_effect')

    with open(driver_path, 'r') as driver_file:
        effect = int(driver_file.read().strip())
        return effect


@endpoint('razer.device.lighting.scroll', 'getScrollBrightness', out_sig='d')
def get_scroll_brightness(self):
    """
    Get the device's brightness

    :return: Brightness
    :rtype: float
    """
    self.logger.debug("DBus call get_scroll_brightness")

    driver_path = self.get_driver_path('scroll_led_brightness')

    with open(driver_path, 'r') as driver_file:
        brightness = round(float(driver_file.read()) * (100.0/255.0), 2)

        return brightness


@endpoint('razer.device.lighting.scroll', 'setScrollBrightness', in_sig='d')
def set_scroll_brightness(self, brightness):
    """
    Set the device's brightness

    :param brightness: Brightness
    :type brightness: int
    """
    self.logger.debug("DBus call set_scroll_brightness")

    driver_path = self.get_driver_path('scroll_led_brightness')

    self.method_args['brightness'] = brightness

    brightness = int(round(brightness * (255.0/100.0)))
    if brightness > 255:
        brightness = 255
    elif brightness < 0:
        brightness = 0

    with open(driver_path, 'w') as driver_file:
        driver_file.write(str(brightness))

    # Notify others
    self.send_effect_event('setBrightness', brightness)


@endpoint('razer.device.lighting.scroll', 'setScrollStatic', in_sig='yyy')
def set_scroll_static(self, red, green, blue):
    """
    Set the device to static colour

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_scroll_static")

    # Notify others
    self.send_effect_event('setStatic', red, green, blue)

    rgb_driver_path = self.get_driver_path('scroll_led_rgb')
    effect_driver_path = self.get_driver_path('scroll_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('0')


@endpoint('razer.device.lighting.scroll', 'setScrollBlinking', in_sig='yyy')
def set_scroll_blinking(self, red, green, blue):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_scroll_pulsate")

    # Notify others
    self.send_effect_event('setPulsate', red, green, blue)

    rgb_driver_path = self.get_driver_path('scroll_led_rgb')
    effect_driver_path = self.get_driver_path('scroll_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('1')


@endpoint('razer.device.lighting.scroll', 'setScrollPulsate', in_sig='yyy')
def set_scroll_pulsate(self, red, green, blue):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_scroll_breathing")

    # Notify others
    self.send_effect_event('setPulsate', red, green, blue)

    rgb_driver_path = self.get_driver_path('scroll_led_rgb')
    effect_driver_path = self.get_driver_path('scroll_led_effect')

    payload = bytes([red, green, blue])

    with open(rgb_driver_path, 'wb') as rgb_driver_file, open(effect_driver_path, 'w') as effect_driver_file:
        rgb_driver_file.write(payload)
        effect_driver_file.write('2')


@endpoint('razer.device.lighting.scroll', 'setScrollSpectrum')
def set_scroll_spectrum(self):
    """
    Set the device to pulsate

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_scroll_spectrum")

    # Notify others
    self.send_effect_event('setSpectrum')

    effect_driver_path = self.get_driver_path('scroll_led_effect')

    with open(effect_driver_path, 'w') as effect_driver_file:
        effect_driver_file.write('4')



"""
Mug class
"""
import re

from openrazer_daemon.hardware.device_base import RazerDeviceBrightnessSuspend as _RazerDeviceBrightnessSuspend


class RazerChromaMugHolder(_RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Chroma Mug Holder
    """
    EVENT_FILE_REGEX = re.compile(r'.*Razer_Chroma_Mug_Holder-if0(1|2)-event-kbd')

    USB_VID = 0x1532
    USB_PID = 0x0F07
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 15]
    DEDICATED_MACRO_KEYS = False
    METHODS = ['get_firmware', 'get_device_name', 'get_device_type_mug', 'has_matrix', 'get_matrix_dims',
               'set_static_effect', 'set_spectrum_effect', 'set_wave_effect', 'set_none_effect', 'set_breath_single_effect', 'set_breath_dual_effect', 'set_breath_random_effect', 'set_blinking_effect',
               'get_brightness', 'set_brightness', 'is_mug_present',
               'set_custom_effect', 'set_key_row']

    RAZER_URLS = {
        "top_img": "https://assets2.razerzone.com/images/mug-holder/e64e507b73e61c44789d996065fd9645-1500x1000mug_01.jpg",
        "side_img": None,
        "perspective_img": "https://assets2.razerzone.com/images/mug-holder/e64e507b73e61c44789d996065fd9645-1500x1000mug_01.jpg"
    }

    def __init__(self, *args, **kwargs):
        super(RazerChromaMugHolder, self).__init__(*args, **kwargs)

    def _close(self):
        super(RazerChromaMugHolder, self)._close()



"""
Macro stuff

Has objects representing key events
Launching programs etc...
"""
import logging
import subprocess
import threading

# pylint: disable=import-error
from openrazer_daemon.keyboard import XTE_MAPPING

# This determins if the macro keys are executed with their natural spacing
XTE_SLEEP = False

class MacroObject(object):
    """
    Macro base object
    """
    def to_dict(self):
        """
        Convert the object to a dict to be sent over DBus

        :return: Dictionary
        :rtype: dict
        """
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, values_dict):
        """
        Create class from dict

        :param values_dict: Dictionary of values (and a type key)
        :type values_dict: dict

        :return: Class object
        """
        del values_dict['type']
        return cls(**values_dict)

class MacroKey(MacroObject):
    """
    Is an object of a key event used in macros
    """
    def __init__(self, key_id, pre_pause, state):
        self.key_id = key_id
        self.pre_pause = pre_pause
        self.state = state

    def __repr__(self):
        return '{0} {1}'.format(self.key_id, self.state)

    def __str__(self):
        return 'MacroKey|{0}|{1}|{2}'.format(self.key_id, self.pre_pause, self.state)

    def to_dict(self):
        return {
            'type': 'MacroKey',
            'key_id': self.key_id,
            'pre_pause': self.pre_pause,
            'state': self.state
        }

    @property
    def xte_key(self):
        """
        Convert key to XTE compatible name

        :return: XTE Name
        :rtype: str
        """
        return XTE_MAPPING.get(self.key_id, self.key_id)

# If it only opens a new tab in chroma - https://askubuntu.com/questions/540939/xdg-open-only-opens-a-new-tab-in-a-new-chromium-window-despite-passing-it-a-url
class MacroURL(MacroObject):
    """
    Is an object of a key event used in macros
    """
    def __init__(self, url):
        self.url = url

    def __repr__(self):
        return '{0}'.format(self.url)

    def __str__(self):
        return 'MacroURL|{0}'.format(self.url)

    def to_dict(self):
        return {
            'type': 'MacroURL',
            'url': self.url,
        }

    def execute(self):
        """
        Open URL in the browser
        """
        proc = subprocess.Popen(['xdg-open', self.url], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        proc.communicate()

class MacroScript(MacroObject):
    """
    Is an object of a key event used in macros
    """
    def __init__(self, script, args=None):
        self.script = script
        if isinstance(args, str):
            self.args = ' ' + args
        else:
            self.args = ''

    def __repr__(self):
        return '{0}'.format(self.script)

    def __str__(self):
        return 'MacroScript|{0}'.format(self.script)

    def to_dict(self):
        return {
            'type': 'MacroScript',
            'script': self.script,
            'args': self.args
        }

    def execute(self):
        """
        Run script
        """
        proc = subprocess.Popen(self.script + self.args, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        proc.communicate()

class MacroRunner(threading.Thread):
    """
    Thread to run macros
    """
    def __init__(self, device_id, macro_bind, macro_data):
        super(MacroRunner, self).__init__()

        self._logger = logging.getLogger('razer.device{0}.macro{1}'.format(device_id, macro_bind))
        self._macro_data = macro_data
        self._macro_bind = macro_bind

    @staticmethod
    def xte_line(key_event):
        """
        Generate a line to be fet into XTE

        :param key_event: Key event object
        :type key_event: MacroKey

        :return: String XTE script
        :rtype: str
        """
        # Save key here to prevent 5odd dictionary lookups
        key = key_event.xte_key

        cmd = ''

        if key is not None:
            if XTE_SLEEP:
                cmd += 'usleep {0}\n'.format(key_event.pre_pause)

            if key_event.state == 'UP':
                cmd += 'keyup {0}\n'.format(key)
            else:
                cmd += 'keydown {0}\n'.format(key)

        return cmd

    def run(self):
        """
        Main thread function
        """

        # TODO move the xte-munging to the init
        xte = ''

        for event in self._macro_data:
            if isinstance(event, MacroKey):
                xte += self.xte_line(event)
            else:
                if xte != '':
                    proc = subprocess.Popen(['xte'], stdin=subprocess.PIPE)
                    proc.communicate(input=xte.encode('ascii'))
                    xte = ''

                # Now run everything else (this just allows for less calls to xte
                if not isinstance(event, MacroKey):
                    event.execute()

        if xte != '':
            proc = subprocess.Popen(['xte'], stdin=subprocess.PIPE)
            proc.communicate(input=xte.encode('ascii'))

        self._logger.debug("Finished running macro %s", self._macro_bind)

def macro_dict_to_obj(macro_dict):
    """
    Converts a macro string to its relevant object

    :param macro_dict: Macro string
    :type macro_dict: dict

    :return: Macro Object
    :rtype: object

    :raises ValueError: When a type isn't known
    """

    # pylint: disable=redefined-variable-type
    if macro_dict['type'] == 'MacroKey':
        result = MacroKey.from_dict(macro_dict)
    elif macro_dict['type'] == 'MacroURL':
        result = MacroURL.from_dict(macro_dict)
    elif macro_dict['type'] == 'MacroScript':
        result = MacroScript.from_dict(macro_dict)
    else:
        raise ValueError("unknown type")

    return result



"""
DBus module

Has an endpoint decorator to wrap a method for DBus
"""
from functools import wraps

def endpoint(interface_name, function_name, in_sig=None, out_sig=None, byte_arrays=False):
    """
    DBus Endpoint

    :param interface_name: DBus Interface name
    :type interface_name: str

    :param function_name: DBus Method name
    :type function_name: str

    :param in_sig: DBus parameter signature
    :type in_sig: str

    :param out_sig: DBus return signature
    :type out_sig: str

    :param byte_arrays: is Byte Array
    :type byte_arrays: bool

    :return: Function
    :rtype: callable
    """
    # pylint: disable=missing-docstring
    def inner_render(func):
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)
        wrapped.endpoint = True
        wrapped.interface = interface_name
        wrapped.name = function_name
        wrapped.in_sig = in_sig
        wrapped.out_sig = out_sig
        wrapped.byte_arrays = byte_arrays
        wrapped.code = func.__code__
        wrapped.globals = func.__globals__
        wrapped.defaults = func.__defaults__
        wrapped.closure = func.__closure__
        return wraps(func)(wrapped)
    return inner_render



import glob
from os.path import dirname, basename, isfile

from ore.core.effects import Effect

_REGISTRY = {}


def register_effect(effect: Effect):
    """
    Registers an effect.

    An effect be of type openrazer_effects.core.effects.Effect to be valid.
    """
    cli_name = effect.get_cli_name(None)
    if cli_name in _REGISTRY:
        raise ValueError("effect {} already registered.".format(cli_name))
    _REGISTRY[cli_name] = effect


def get_registry():
    _load_all_submodules()
    return _REGISTRY


def _load_all_submodules():
    import ore.effects
    modules = glob.glob(dirname(ore.effects.__file__) + "/*.py")
    files = ["ore.effects.{}".format(basename(f)[:-3]) for f in modules if isfile(f) and not f.endswith('__init__.py')]
    f = list(map(__import__, files))
    return f



"""
Module to handle custom colours
"""

import pgi
pgi.install_as_gi()

import gi
gi.require_version('Gdk', '3.0')
from gi.repository import Gdk
import struct
import subprocess


KEY_MAPPING = {
    # Row 0
    'ESC': (0, 1), 'F1': (0, 3), 'F2': (0, 4), 'F3': (0, 5), 'F4': (0, 6), 'F5': (0, 7), 'F6': (0, 8), 'F7': (0, 9), 'F8': (0, 10), 'F9': (0, 11), 'F10': (0, 12), 'F11': (0, 13), 'F12': (0, 14), 'PRTSCR': (0, 15), 'SCRLK': (0, 16), 'PAUSE': (0, 17), 'LOGO': (0, 20), 'JP1': (0, 21),
    # Row 1
    'M1': (1, 0), 'BACKTICK': (1, 1), '1': (1, 2), '2': (1, 3), '3': (1, 4), '4': (1, 5), '5': (1, 6), '6': (1, 7), '7': (1, 8), '8': (1, 9), '9': (1, 10), '0': (1, 11), 'DASH': (1, 12), 'EQUALS': (1, 13), 'BACKSPACE': (1, 14), 'INS': (1, 15), 'HOME': (1, 16), 'PAGEUP': (1, 17), 'NUMLK': (1, 18), 'NPFORWARDSLASH': (1, 19), 'NPASTERISK': (1, 20), 'NPDASH': (1, 21),
    # Row 2
    'M2': (2, 0), 'TAB': (2, 1), 'Q': (2, 2), 'W': (2, 3), 'E': (2, 4), 'R': (2, 5), 'T': (2, 6), 'Y': (2, 7), 'U': (2, 8), 'I': (2, 9), 'O': (2, 10), 'P': (2, 11), 'LEFTSQUAREBRACKET': (2, 12), 'RIGHTSQUAREBRACKET': (2, 13), 'DELETE': (2, 15), 'END': (2, 16), 'PAGEDOWN': (2, 17), 'NP7': (2, 18), 'NP8': (2, 19), 'NP9': (2, 20), 'NPPLUS': (2, 21),
    # Row 3
    'M3': (3, 0), 'CAPSLK': (3, 1), 'A': (3, 2), 'S': (3, 3), 'D': (3, 4), 'F': (3, 5), 'G': (3, 6), 'H': (3, 7), 'J': (3, 8), 'K': (3, 9), 'L': (3, 10), 'SEMICOLON': (3, 11), 'APOSTROPHE': (3, 12), 'POUNDSIGN': (3, 13), 'RETURN': (3, 14), 'NP4': (3, 18), 'NP5': (3, 19), 'NP6': (3, 20),
    # Row 4
    'M4': (4, 0), 'LEFTSHIFT': (4, 1), 'BACKSLASH': (4, 2), 'Z': (4, 3), 'X': (4, 4), 'C': (4, 5), 'V': (4, 6), 'B': (4, 7), 'N': (4, 8), 'M': (4, 9), 'COMMA': (4, 10), 'PERIOD': (4, 11), 'FORWARDSLASH': (4, 12), 'JP2': (4, 13), 'RIGHTSHIFT': (4, 14), 'UPARROW': (4, 16), 'NP1': (4, 18), 'NP2': (4, 19), 'NP3': (4, 20), 'ENTER': (4, 21),
    # Row 5
    'M5': (5, 0), 'LEFTCTRL': (5, 1), 'SUPER': (5, 2), 'LEFTALT': (5, 3), 'SPACE': (5, 7), 'RIGHTALT': (5, 11), 'FN': (5, 12), 'CTXMENU': (5, 13), 'RIGHTCTRL': (5, 14), 'LEFTARROW': (5, 15), 'DOWNARROW': (5, 16), 'RIGHTARROW': (5, 17), 'NP0': (5, 19), 'NPPERIOD': (5, 20),

    # Additional mappings
    'MACROMODE': (0, 11), 'GAMEMODE': (0, 12), 'MUTE': (0, 3), 'VOL_DOWN': (0, 4), 'VOL_UP': (0, 5), 'MEDIA_BACK': (0, 7), 'MEDIA_PLAY': (0, 8), 'MEDIA_FORWARD': (0, 9), 'BRIGHTNESSDOWN': (0, 13), 'BRIGHTNESSUP': (0, 14)
}


#      0   1   2   3   4    5   6   7
#    ------------------------------------
# 0 |  1   2   3   4   5        UP
# 1 |  6   7   8   9   10  LEFT    RIGHT
# 2 |  11  12  13  14  15  TMB DOWN MS
#
#
#

TARTARUS_KEY_MAPPING = {
    '1': (0, 0),
    '2': (0, 1),
    '3': (0, 2),
    '4': (0, 3),
    '5': (0, 4),
    'UP': (0, 6),

    '6': (1, 0),
    '7': (1, 1),
    '8': (1, 2),
    '9': (1, 3),
    '10': (1, 4),
    'LEFT': (1, 5),
    'RIGHT': (1, 7),

    '11': (2, 0),
    '12': (2, 1),
    '13': (2, 2),
    '14': (2, 3),
    '15': (2, 4),
    'THUMB': (2, 5),
    'DOWN': (2, 6),
    'MODE_SWITCH': (2, 7)

}

#      0   1   2   3   4    5   6   7
#    ------------------------------------
# 0 |  1   2   3   4   5        UP
# 1 |  6   7   8   9   10  LEFT    RIGHT
# 2 |  11  12  13  14  15  TMB DOWN MS
# 3 |  16  17  18  19  20
#

ORBWEAVER_KEY_MAPPING = {
    '1': (0, 0),
    '2': (0, 1),
    '3': (0, 2),
    '4': (0, 3),
    '5': (0, 4),
    'UP': (0, 6),

    '6': (1, 0),
    '7': (1, 1),
    '8': (1, 2),
    '9': (1, 3),
    '10': (1, 4),
    'LEFT': (1, 5),
    'RIGHT': (1, 7),

    '11': (2, 0),
    '12': (2, 1),
    '13': (2, 2),
    '14': (2, 3),
    '15': (2, 4),
    'THUMB': (2, 5),
    'DOWN': (2, 6),
    'MODE_SWITCH': (2, 7),

    '16': (3, 0),
    '17': (3, 1),
    '18': (3, 2),
    '19': (3, 3),
    '20': (3, 4)
}

# Naga hex buttons are normal 1-7 keys
#      0 1 2 3 4
#   --------------
# 0 |    1   2
# 1 |  4       3
# 2 |    5   7
# 3 |      6
NAGA_HEX_V2_KEY_MAPPING = {
    '1': (0, 1), '2': (0, 3),
    '4': (1, 0), '3': (1, 4),
    '5': (2, 1), '7': (2, 3),
    '6': (3, 2)
}

#
EVENT_MAPPING = {
    1: 'ESC', 2: '1',  3: '2',  4: '3',  5: '4',  6: '5',  7: '6',  8: '7',  9: '8',
    10: '9', 11: '0', 12: 'DASH', 13: 'EQUALS', 14: 'BACKSPACE', 15: 'TAB', 16: 'Q', 17: 'W', 18: 'E', 19: 'R',
    20: 'T', 21: 'Y', 22: 'U', 23: 'I', 24: 'O', 25: 'P', 26: 'LEFTSQUAREBRACKET', 27: 'RIGHTSQUAREBRACKET', 28: 'RETURN', 29: 'LEFTCTRL',
    30: 'A', 31: 'S', 32: 'D', 33: 'F', 34: 'G', 35: 'H', 36: 'J', 37: 'K', 38: 'L', 39: 'SEMICOLON',
    40: 'APOSTROPHE', 41: 'BACKTICK', 42: 'LEFTSHIFT', 43: 'POUNDSIGN', 44: 'Z', 45: 'X', 46: 'C', 47: 'V', 48: 'B', 49: 'N',
    50: 'M', 51: 'COMMA', 52: 'PERIOD', 53: 'FORWARDSLASH', 54: 'RIGHTSHIFT', 55: 'NPASTERISK', 56: 'LEFTALT', 57: 'SPACE',
    58: 'CAPSLK', 59: 'F1',
    60: 'F2', 61: 'F3', 62: 'F4', 63: 'F5', 64: 'F6', 65: 'F7', 66: 'F8', 67: 'F9', 68: 'F10', 69: 'NUMLK',
    70: 'SCRLK', 71: 'NP7', 72: 'NP8', 73: 'NP9', 74: 'NPDASH', 75: 'NP4', 76: 'NP5', 77: 'NP6', 78: 'NPPLUS', 79: 'NP1',
    80: 'NP2', 81: 'NP3', 82: 'NP0', 83: 'NPPERIOD', 86: 'BACKSLASH', 87: 'F11', 88: 'F12',
    96: 'ENTER', 97: 'RIGHTCTRL', 98: 'NPFORWARDSLASH', 99: 'PRTSCR',
    100: 'RIGHTALT', 102: 'HOME', 103: 'UPARROW', 104: 'PAGEUP', 105: 'LEFTARROW', 106: 'RIGHTARROW', 107: 'END', 108: 'DOWNARROW', 109: 'PAGEDOWN', 110: 'INS',
    111: 'DELETE', 113: 'MUTE', 114: 'VOL_DOWN', 115: 'VOL_UP', 119: 'PAUSE',
    125: 'SUPER', 127: 'CTXMENU',
    163: 'MEDIA_FORWARD', 164: 'MEDIA_PLAY', 165: 'MEDIA_BACK',
    183: 'M1', 184: 'M2', 185: 'M3', 186: 'M4', 187: 'M5', 188: 'MACROMODE', 189: 'GAMEMODE', 190: 'BRIGHTNESSDOWN', 194: 'BRIGHTNESSUP'
}

TARTARUS_EVENT_MAPPING = {
    15: '1',
    16: '2',
    17: '3',
    18: '4',
    19: '5',
    58: '6',
    30: '7',
    31: '8',
    32: '9',
    33: '10',
    42: '11',
    44: '12',
    45: '13',
    46: '14',
    47: '15',
    56: 'MODE_SWITCH',
    57: 'THUMB',
    103: 'UP',
    105: 'LEFT',
    106: 'RIGHT',
    108: 'DOWN',
}

ORBWEAVER_EVENT_MAPPING = {
    41: '1',
    2: '2',
    3: '3',
    4: '4',
    5: '5',
    15: '6',
    16: '7',
    17: '8',
    18: '9',
    19: '10',
    58: '11',
    30: '12',
    31: '13',
    32: '14',
    33: '15',
    42: '16',
    44: '17',
    45: '18',
    46: '19',
    47: '20',
    56: 'MODE_SWITCH',
    57: 'THUMB',
    103: 'UP',
    105: 'LEFT',
    106: 'RIGHT',
    108: 'DOWN',
}

NAGA_HEX_V2_EVENT_MAPPING = {
    2: '1',  3: '2',  4: '3',  5: '4',  6: '5',  7: '6',  8: '7'
}

XTE_MAPPING = {
    'ESC': 'Escape',
    'DASH': 'minus',
    'EQUALS': 'equal',
    'BACKSPACE': 'BackSpace',
    'TAB': 'Tab',
    'LEFTSQUAREBRACKET': 'bracketleft',
    'RIGHTSQUAREBRACKET': 'bracketright',
    'RETURN': 'Return',
    'LEFTCTRL': 'Control_L',
    'SEMICOLON': 'semicolon',
    'APOSTROPHE': 'apostrophe',
    'BACKTICK': 'grave',
    'LEFTSHIFT': 'Shift_L',
    'POUNDSIGN': 'numbersign',
    'COMMA': 'comma',
    'PERIOD': 'period',
    'FORWARDSLASH': 'slash',
    'RIGHTSHIFT': 'Shift_R',
    'NPASTERISK': 'KP_Multiply',
    'LEFTALT': 'Alt_L',
    'SPACE': 'space',
    'CAPSLK': 'Caps_Lock',
    'NUMLK': 'Num_Lock',
    'SCRLK': 'Scroll_Lock',
    'PAUSE': 'Pause',
    'NP7': '7',
    'NP8': '8',
    'NP9': '9',
    'NPDASH': 'KP_Subtract',
    'NP4': '4',
    'NP5': '5',
    'NP6': '6',
    'NPPLUS': 'KP_Add',
    'NP1': '1',
    'NP2': '2',
    'NP3': '3',
    'NP0': '0',
    'PRTSCR': 'Print',
    'NPPERIOD': 'KP_Decimal',
    'BACKSLASH': 'backslash',
    'ENTER': 'Enter',
    'RIGHTCTRL': 'Control_r',
    'NPFORWARDSLASH': 'KP_Divide',
    'RIGHTALT': 'Alt_R',
    'HOME': 'Home',
    'UPARROW': 'Up',
    'PAGEUP': 'Page_Up',
    'LEFTARROW': 'Left',
    'RIGHTARROW': 'Right',
    'END': 'End',
    'DOWNARROW': 'Down',
    'PAGEDOWN': 'Page_Down',
    'INS': 'Insert',
    'DELETE': 'Delete',
    'SUPER': 'Super_L',
    'CTXMENU': 'Menu',
    'M1': 'XF86Tools',
    'M2': 'XF86Launch5',
    'M3': 'XF86Launch6',
    'M4': 'XF86Launch7',
    'M5': 'XF86Launch8',
    'FN': None,
    'GAMEMODE': None,
    'MACROMODE': None,
}


class KeyDoesNotExistError(Exception):
    """
    Simple custom error
    """
    pass


class NoBackupError(Exception):
    pass


class RGB(object):

    @staticmethod
    def clamp(value):
        """
        Clamp a value to 0-255

        :param value: Value to be clamped
        :type value: integer or float

        :return: Integer in the range of 0-255
        :rtype: int
        """
        result = int(value)

        if value > 255:
            result = 255
        elif value < 0:
            result = 0

        return result

    def __init__(self, red=0, green=0, blue=0):
        self._red = red
        self._green = green
        self._blue = blue

    @property
    def red(self):
        """
        Getter for red element

        :return: Red element
        :rtype: int
        """
        return self._red

    @red.setter
    def red(self, value):
        """
        Setter for red value

        :param value: Red value
        :type value: int or float
        """
        self._red = RGB.clamp(value)

    @property
    def green(self):
        """
        Getter for green element

        :return: Green element
        :rtype: int
        """
        return self._green

    @green.setter
    def green(self, value):
        """
        Setter for green value

        :param value: Green value
        :type value: int or float
        """
        self._green = RGB.clamp(value)

    @property
    def blue(self):
        """
        Getter for blue element

        :return: Blue element
        :rtype: int
        """
        return self._blue

    @blue.setter
    def blue(self, value):
        """
        Setter for blue value

        :param value: Blue value
        :type value: int or float
        """
        self._blue = RGB.clamp(value)

    def set(self, colour_tuple):
        """
        Sets all the colours at once

        :param colour_tuple: Tuple of R,G,B elements
        :type colour_tuple: tuple
        """
        # Shortcut to clamp all parameters and assign to the 3 variables
        self._red, self._green, self._blue = list(map(RGB.clamp, colour_tuple))

    def get(self):
        """
        Gets all the colours as a tuple

        :return: RGB tuple
        :rtype: tuple
        """
        return self._red, self._green, self._blue

    def __bytes__(self):
        """
        Convert to bytes

        :return: Byte string
        :rtype: bytearray
        """
        return bytes((self._red, self._green, self._blue))

    def __repr__(self):
        """
        String representation

        :return: String
        :rtype: str
        """
        return "RGB Object (#{0:02X}{1:02X}{2:02X})".format(self._red, self._green, self._blue)


class KeyboardColour(object):
    """
    Keyboard class which represents the colour state of the keyboard.
    """

    @staticmethod
    def gdk_colour_to_rgb(gdk_color):
        """
        Converts GDK colour to (R,G,B) tuple

        :param gdk_color: GDK colour
        :type gdk_color: Gdk.Color or tuple

        :return: Tuple of 3 ints
        :rtype: tuple
        """
        if isinstance(gdk_color, (list, tuple)):
            return gdk_color

        assert type(gdk_color) is Gdk.Color, "Is not of type Gdk.Color"

        red = int(gdk_color.red_float * 255)
        green = int(gdk_color.green_float * 255)
        blue = int(gdk_color.blue_float * 255)

        return red, green, blue

    def __init__(self):
        self.rows = []

        self.backup = None

        self.reset_rows()

    def backup_configuration(self):
        """
        Backs up the current configuration
        """
        self.backup = KeyboardColour()
        self.backup.get_from_total_binary(self.get_total_binary())

    def restore_configuration(self):
        """
        Restores the previous configuration
        """
        if self.backup is None:
            raise NoBackupError()

        self.rows = self.backup.rows
        self.backup = None

    def get_rows_raw(self):
        """
        Gets the raw representation of the rows

        :return: Rows
        :rtype: list
        """
        return self.rows

    def reset_rows(self):
        """
        Reset the rows of the keyboard
        """
        self.rows.clear()

        for row in range(0, 6):
            # Create 22 rgb values
            self.rows.append([RGB() for _ in range(0,22)])

    def set_key_colour(self, row, col, colour):
        """
        Set the colour of a key

        :param row: Row ID
        :type row: int

        :param col: Column ID
        :type col: int

        :param colour: Colour to set
        :type colour: Gdk.Color or tuple

        :raises KeyDoesNotExistError: If given key does not exist
        """
        self.rows[row][col].set(KeyboardColour.gdk_colour_to_rgb(colour))

    def get_key_colour(self, key):
        """
        Get the colour of a key

        :param key: Key to set the colour of
        :type key: str

        :raises KeyDoesNotExistError: If given key does not exist
        """
        if key not in KEY_MAPPING:
            raise KeyDoesNotExistError("The key \"{0}\" does not exist".format(key))

        row_id, col_id = KEY_MAPPING[key]
        return self.rows[row_id][col_id].get()

    def reset_key(self, row, col):
        """
        Reset the colour of a key

        :param row: Row ID
        :type row: int

        :param col: Column ID
        :type col: int

        :raises KeyDoesNotExistError: If given key does not exist
        """
        self.rows[row][col].set((0, 0, 0))

    def get_row_binary(self, row_id):
        """
        Gets the binary payload for a given row

        :param row_id: Row ID
        :type row_id: int

        :return: Byte string of 67 bytes, Row ID byte then 22 RGB bytes
        :rtype: bytearray
        """
        assert isinstance(row_id, int), "Row ID is not an int"

        payload = bytes([row_id, 0x00, len(self.rows[row_id])-1])

        for rgb in self.rows[row_id]:
            payload += bytes(rgb)

        return payload

    def get_total_binary(self):
        """
        Gets the binary payload for the whole keyboard

        :return: Byte string of 6*67 bytes, (Row ID byte then 22 RGB bytes) * 6
        :rtype: bytearray
        """
        payload = b''

        for row in range(0,6):
            payload += self.get_row_binary(row)

        return payload

    def get_from_total_binary(self, binary_blob):
        """
        Load in a binary blob which is the output from get_total_binary

        :param binary_blob: Binary blob
        :type binary_blob: bytearray
        """
        self.reset_rows()

        for row_id in range (0, 6):
            binary_blob = binary_blob[1:] # Skip first byte

            for col_id, binary_rgb in enumerate([binary_blob[i:i+3] for i in range(0, 66, 3)]):
                rgb = struct.unpack('=BBB', binary_rgb)
                self.rows[row_id][col_id].set(rgb)

            binary_blob = binary_blob[66:] # Skip the current row


def get_keyboard_layout():
    """
    Function to get the keyboard layout

    I can see this becoming an ugly mess as it'll need to support multiple keyboard layouts which have different names
    on different systems

    :return: Keyboard layout
    :rtype: str
    """
    cmd = ['setxkbmap', '-query']
    output = subprocess.check_output(cmd)

    result = 'gb'

    if output:
        output = output.decode('utf-8').splitlines()
        layout = 'gb'
        variant = ''
        for line in output:
            if line.startswith('layout'):
                layout = line.split(':', 1)[1].strip()
                if layout.find(',') > -1:
                    layout = layout.split(',')[0]

            elif line.startswith('variant'):
                variant = line.split(':', 1)[1].strip().split(',')[0]

                if 'latin9' in variant: # Removes some rubbish from ubuntu
                    variant = 'latin9'

        if variant == '':
            result = layout
        else:
            result = layout + '-' + variant

    # If the user has an international layout variant ignore that part
    result = result.replace('-altgr-intl', '')

    return result



"""
BlackWidow Macro accessors
"""
from openrazer_daemon.dbus_services import endpoint


@endpoint('razer.device.macro', 'getMacros', out_sig='s')
def get_macros(self):
    """
    Get macros

    :return: JSON of macros
    :rtype: str
    """
    self.logger.debug("DBus call get_macros")

    return self.key_manager.dbus_get_macros()


@endpoint('razer.device.macro', 'deleteMacro', in_sig='s')
def delete_macro(self, macro_key):
    """
    Delete macro from key

    :param macro_key: Macro key to delete bound macro from
    :type macro_key: str
    """
    self.logger.debug("DBus call delete_macro")

    self.key_manager.dbus_delete_macro(macro_key)


@endpoint('razer.device.macro', 'addMacro', in_sig='ss')
def add_macro(self, macro_bind_key, macro_json):
    """
    Add macro to key

    The macro_json should be JSON form of a list of dictionaries
    :param macro_bind_key: Macro key to delete bound macro from
    :type macro_bind_key: str

    :param macro_json: JSON list
    :type macro_json: str
    """
    self.logger.debug("DBus call add_macro")

    self.key_manager.dbus_add_macro(macro_bind_key, macro_json)



class Effect(object):
    """
    Effect interface.

    All Effects are subject to this interface.
    """

    def __init__(self, **kwargs):
        """
        Constructor must accepts kwargs.

        audio_device is passed in to connect with local audio devices.
        """
        pass

    def get_name(self):
        """
        The user-friendly name of this effect.
        """
        raise NotImplementedError()

    def get_description(self):
        """
        The user-friendly, one-line description of this effect.
        """
        return None

    def get_cli_name(self):
        """
        The cli-friendly slug of this effect.
        """
        raise NotImplementedError()

    def get_author(self):
        """
        The author of this effect.
        """
        raise NotImplementedError()

    def get_effect_type(self):
        """
        The type of this effect. Either "keyboard" or "mouse"
        """
        raise NotImplementedError()

    def start(self):
        """
        Begin this effect.
        """
        raise NotImplementedError()


class KeyboardEffect(Effect):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_effect_type(self):
        return "keyboard"


class MouseEffect(Effect):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_effect_type(self):
        return "mouse"



#!/usr/bin/env python3.6
import math
import random
import typing

from openrazer.client.devices.keyboard import RazerKeyboard
from openrazer.client.fx import Frame
from ore.core.program import KeyboardProgram
from ore.core.utils import ROWS, COLS
from ore.registry.registry import register_effect


def deg_to_rad(theta):
    return theta * math.pi / 180


class Explosion(object):
    def __init__(self, center=(ROWS // 2, COLS // 2), radius=0, color=(255, 75, 0), resolution=20):
        self.center = center
        self.radius = radius
        self.resolution = resolution
        self.color = color

    def expand(self, delta=0.45):
        self.radius += delta

    def write_to(self, matrix: Frame) -> bool:
        delta = 360 / self.resolution
        row, col = self.center

        drawn = False

        theta = 0
        while theta <= 360:

            rad = deg_to_rad(theta)

            dx = int(col + math.cos(rad) * self.radius)
            dy = int(row + math.sin(rad) * self.radius)

            if 0 <= dx < COLS and 0 <= dy < ROWS:
                matrix[dy, dx] = self.color
                drawn = True
            theta += delta

        return drawn


@register_effect
class ExplosionEffect(KeyboardProgram):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.num_explosions = 3
        self.explosions = self.make_explosions()
        self.clear_every_frame = False

    def get_name(self):
        return "Explosions"

    def get_description(self):
        return "Light and dark explosions that compliment each other."

    def get_cli_name(self):
        return "explosions"

    def get_author(self):
        return "explodes"

    def make_explosions(self, darken=1):
        return [self.make_explosion(darken=darken) for _ in range(self.num_explosions)]

    def make_explosion(self, darken=1):
        col = random.randint(0, COLS - 1)
        row = random.randint(0, ROWS - 1)
        # col = COLS // 2
        # row = ROWS // 2
        color = (
            random.randint(50, 255) // darken,
            random.randint(50, 255) // darken,
            random.randint(50, 255) // darken
        )
        return Explosion(center=(row, col), color=color)

    def init(self, kb: RazerKeyboard):
        kb.brightness = 100

    def draw(self, kb: RazerKeyboard, matrix: Frame, frame_num: int, dt: float) -> typing.Any:

        if frame_num % 15 == 0:
            # matrix.reset()
            self.explosions = self.make_explosions(darken=(frame_num % 2) * 15 + 1)

        for index, explosion in enumerate(self.explosions):
            explosion.expand()
            if not explosion.write_to(matrix):
                self.explosions[index] = self.make_explosion()


if __name__ == '__main__':
    effect = ExplosionEffect()
    effect.start()



import struct
import threading
from queue import Queue

import numpy
import pyaudio


class DeviceFinder(object):
    def __init__(self, p):
        self.p = p

    def find_output_devices(self):
        """
        Find devices suitable for writing.
        """
        for device in self.devicesiter():
            if device["maxOutputChannels"] > 0:
                yield device

    def find_input_devices(self):
        """
        Find devices suitable for recording.
        """
        for device in self.devicesiter():
            if device["maxInputChannels"] > 0:
                yield device

    def find_named_device(self, name):
        for device in self.devicesiter():
            if device["name"] == name:
                return device
        raise ValueError("device {} not found".format(name))

    def devicesiter(self):
        count = self.p.get_device_count()
        for index in range(count):
            device = self.p.get_device_info_by_index(index)
            yield device


class DeviceReader(object):

    def __init__(self, p, device):
        self.p = p
        self.device = device

    def stream_output(self, buffer_size=2 ** 10):
        return self.create_stream(is_input=False, buffer_size=buffer_size)

    def stream_input(self, buffer_size=2 ** 10):
        return self.create_stream(is_input=True, buffer_size=buffer_size)

    def create_stream(self, is_input, buffer_size):
        stream = AudioStream(self.device)

        thread = threading.Thread(
            name="device-thread-{}".format(self.device["name"]),
            target=self._start_stream,
            args=(stream, is_input, buffer_size),
            daemon=True)
        thread.start()

        return stream

    def _generate_null_stream(self, stream):
        import time

        while True:
            stream(b"\x00" * 304, 0, 0, 0)
            time.sleep(0.2)

    def _start_stream(self, stream, is_input, buffer_size):
        self.p.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=int(self.device.get("defaultSampleRate", 44100)),
            input=is_input,
            output=not is_input,
            frames_per_buffer=buffer_size,
            input_device_index=self.device["index"],
            stream_callback=stream)


class AudioStream(object):
    flags = {
        pyaudio.paInputUnderflow: "paInputUnderflow",
        pyaudio.paInputOverflow: "paInputOverflow",
        pyaudio.paOutputUnderflow: "paOutputUnderflow",
        pyaudio.paOutputOverflow: "paOutputOverflow",
        pyaudio.paPrimingOutput: "paPrimingOutput",
    }

    def __init__(self, device):
        self.device = device
        self.closed = False
        self.queue = Queue()

    def __call__(self, in_data, frame_count, time_info, status):
        if status == 0:
            self.queue.put(in_data, block=True)
        return None, pyaudio.paComplete if self.closed else pyaudio.paContinue

    def __iter__(self):
        while True:
            value = self.queue.get(block=True)
            if self.closed:
                break
            yield value

    def close(self):
        self.closed = True


class AudioAnalyzer(object):

    def __init__(self, buckets=6, scale=10, exponent=1):
        self.buckets = buckets
        self.scale = scale
        self.exponent = exponent

    def calculate_levels(self, data):
        # Use FFT to calculate volume for each frequency

        # Convert raw sound data to Numpy array
        fmt = "%dH" % (len(data) // 2)
        data2 = struct.unpack(fmt, data)
        data2 = numpy.array(data2, dtype='h')

        # Apply FFT
        fourier = numpy.fft.fft(data2)
        ffty = numpy.abs(fourier[0:len(fourier) // 2]) / 1000
        ffty1 = ffty[:len(ffty) // 2]
        ffty2 = ffty[len(ffty) // 2::] + 2
        ffty2 = ffty2[::-1]
        ffty = ffty1 + ffty2
        ffty = numpy.log(ffty) - 2

        fourier = list(ffty)[4:-4]
        fourier = fourier[:len(fourier) // 2]

        size = len(fourier)

        # Add up for 6 lights
        levels = [sum(fourier[i:(i + size // self.buckets)]) for i in range(0, size, size // self.buckets)][
                 :self.buckets]

        return self._normalize_levels(levels)

    def _normalize_levels(self, levels):
        out = [0] * len(levels)
        lo = min(levels)
        hi = max(levels)
        if lo == hi:
            return out
        for index, level in enumerate(levels):
            p = (level - lo) / (hi - lo)
            out[index] = p ** self.exponent
        return out


class Histogram(object):
    BOXES = ('\u2581', '\u2582', '\u2583', '\u2584', '\u2585', '\u2586', '\u2587', '\u2588')

    def __init__(self, size):
        self.size = size
        self.values = [0] * size

    def set_values(self, values):
        self.values = values

    def _char(self, index):
        percentage = self.values[index]
        char_index = int((len(Histogram.BOXES) - 1) * percentage)
        return Histogram.BOXES[char_index]

    def __str__(self):
        return u"".join(self._char(index) for index in range(self.size))


class DeviceGraph(object):

    def __init__(self, p, device_name, buffer_size=2 ** 11, buckets=64):
        self.device_name = device_name

        finder = DeviceFinder(p)
        device = finder.find_named_device(device_name)
        self.analyzer = AudioAnalyzer(buckets=buckets)
        self.reader = DeviceReader(p, device)
        self.histogram = Histogram(buckets)
        self.stream = self.reader.stream_input(buffer_size=buffer_size)
        self.stream_data = iter(self.stream)

    def read_row(self):
        try:
            data = next(self.stream_data)
        except StopIteration:
            return
        else:
            levels = self.analyzer.calculate_levels(data)
            self.histogram.set_values(levels)

    def close(self):
        self.stream.close()


def counter(initial=0, step=1):
    i = initial
    while True:
        yield i
        i += step


def graph_all_devices(p, device_names=None, buffer_size=2 ** 11, buckets=32):
    p = pyaudio.PyAudio()
    finder = DeviceFinder(p)

    if device_names is None:
        device_names = [device["name"] for device in finder.find_input_devices() if " " not in device["name"]]
        print("have devices: {}".format(", ".join(device_names)))

    graphs = [DeviceGraph(
        p,
        device_name,
        buffer_size=buffer_size,
        buckets=buckets)
        for device_name in device_names]

    for index in counter():
        print("{:>5}".format(index), end=" :: ")
        for graph in graphs:
            graph.read_row()
            print("{}: {}".format(graph.device_name, graph.histogram), end=" ")
        print("\n", end="")


if __name__ == "__main__":
    p = pyaudio.PyAudio()
    graph_all_devices(p)



"""
Naga Hex DPI
"""
import struct
from openrazer_daemon.dbus_services import endpoint

@endpoint('razer.device.dpi', 'setDPI', in_sig='qq')
def set_dpi_xy_byte(self, dpi_x, dpi_y):
    """
    Set the DPI on the mouse, Takes in 4 bytes big-endian and converts it to bytes

    :param dpi_x: X DPI
    :type dpi_x: int
    :param dpi_y: Y DPI
    :type dpi_x: int
    """
    self.logger.debug("DBus call set_dpi_both")

    driver_path = self.get_driver_path('dpi')

    if dpi_x > 6750:
        dpi_x = 6750
    elif dpi_x < 100:
        dpi_x = 100
    if dpi_y > 6750:
        dpi_y = 6750
    elif dpi_y < 100:
        dpi_y = 100

    dpi_x_scaled = int(round(dpi_x / 6750 * 255, 2))
    dpi_y_scaled = int(round(dpi_y / 6750 * 255, 2))

    dpi_bytes = struct.pack('>BB', dpi_x_scaled, dpi_y_scaled)

    with open(driver_path, 'wb') as driver_file:
        driver_file.write(dpi_bytes)


@endpoint('razer.device.dpi', 'getDPI', out_sig='ai')
def get_dpi_xy_byte(self):
    """
    get the DPI on the mouse

    :return: List of X, Y DPI
    :rtype: list of int
    """
    self.logger.debug("DBus call get_dpi_both")

    driver_path = self.get_driver_path('dpi')

    with open(driver_path, 'r') as driver_file:
        result = driver_file.read()
        dpi_x, dpi_y = [int(dpi) for dpi in result.strip().split(':')]

    dpi_x = int(round(dpi_x / 255 * 6750, 2))
    dpi_y = int(round(dpi_y / 255 * 6750, 2))

    return [dpi_x, dpi_y]



import types
import inspect


def get_attrs(obj):
    all_obj_attrs = [x for x in dir(obj) if not x.startswith('_')]
    props = []
    funcs = []
    fields = []

    for obj_attr in all_obj_attrs:
        objs_to_check = list(obj.__class__.__mro__)
        objs_to_check.insert(0, obj)

        for obj_class in objs_to_check:
            try:
                attr = getattr(obj_class, obj_attr)

                if isinstance(attr, property):
                    get_sig = str(inspect.signature(attr.fget))
                    if '->' in get_sig:
                        get_sig = ':' + get_sig.split('-> ')[1]
                    else:
                        get_sig = ''

                    if attr.fset is not None:
                        set_sig = inspect.signature(attr.fset)
                        if len(set_sig.parameters.keys()) > 0:
                            set_type = str(set_sig.parameters[list(set_sig.parameters.keys())[-1]])
                        else:
                            set_type = ''

                        if ':' in set_type:
                            set_type = ':' + set_type.split(':')[1]
                        else:
                            set_type = ''
                        props.append(obj_attr + ' - get{0}, set{1}'.format(get_sig, set_type))
                    else:
                        props.append(obj_attr + ' - get{0}'.format(get_sig))
                    break
                elif isinstance(attr, types.FunctionType):
                    funcs.append(obj_attr + str(inspect.signature(attr)))
                    break
            except Exception as err:
                pass
        else:
            fields.append(obj_attr)

    return sorted(props), sorted(funcs), sorted(fields)


def print_attrs(obj, recurse_to=None, indent=0):
    if recurse_to is None:
        recurse_to = set()
    else:
        recurse_to = set(recurse_to)

    props, funcs, fields = get_attrs(obj)

    print2 = lambda x: print((" " * indent) + x)

    print(obj.__class__.__name__)
    if len(props) > 0:
        if len(funcs) + len(fields) == 0:
            print2(" └ properties:")
            pipe = ' '
        else:
            print2(" ├ properties:")
            pipe = '│'

        for index, prop in enumerate(props, start=1):
            if prop in recurse_to:
                if len(props) != index:
                    print((" " * indent) + ' {0}  ├- {1} - '.format(pipe, prop), end='')
                    print_attrs(getattr(obj, prop), recurse_to=(recurse_to - set(prop)), indent=indent + 7)
                else:
                    print((" " * indent) + ' {0}  └- {1} - '.format(pipe, prop), end='')
                    print_attrs(getattr(obj, prop), recurse_to=(recurse_to - set(prop)), indent=indent + 7)
            else:
                if len(props) != index:
                    print2(' {0}  ├- {1}'.format(pipe, prop))
                else:
                    print2(' {0}  └- {1}'.format(pipe, prop))
    if len(funcs) > 0:
        if len(fields) == 0:
            print2(" └ methods:")
            pipe = ' '
        else:
            print2(" ├ methods:")
            pipe = '│'

        for index, func in enumerate(funcs, start=1):
            if len(funcs) != index:
                print2(' {0}  ├- {1}'.format(pipe, func))
            else:
                print2(' {0}  └- {1}'.format(pipe, func))
    if len(fields) > 0:
        print2(" └ fields:")

        for index, field in enumerate(fields, start=1):
            if field in recurse_to:
                if len(fields) != index:
                    print((" " * indent) + '   ├- {0} - '.format(field), end='')
                    print_attrs(getattr(obj, field), recurse_to=(recurse_to - set(field)), indent=indent + 6)
                else:
                    print((" " * indent) + '   └- {0} - '.format(field), end='')
                    print_attrs(getattr(obj, field), recurse_to=(recurse_to - set(field)), indent=indent + 6)
            else:
                if len(fields) != index:
                    print2('   ├- {0}'.format(field))
                else:
                    print2('   └- {0}'.format(field))



#!/usr/bin/env python

import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported')

setup(
    name="openrazer_effects",
    description="Effects for OpenRazer",
    url="https://github.com/explodes/openrazer_effects",
    author="Evan Leis",
    version="1.0.2",
    packages=find_packages(".", exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        "dbus-python==1.2.4",
        "numpy==1.13.3",
        "pgi==0.0.11.1",
        "PyAudio==0.2.11",
        "six==1.11.0",
    ],
    scripts=["openrazer_effects"],
    license="GPLv2",
    setup_requires=[
        "pytest-runner==3.0"
    ],
    tests_require=[
        "pytest==3.2.3",
    ],
    python_requires=">=3.6",
    zip_safe=True,
)



"""
Contains functionality specific to keyboard-like devices
"""
import dbus as _dbus

from openrazer.client.constants import MACRO_LED_STATIC, MACRO_LED_BLINK
from openrazer.client.devices import RazerDevice as __RazerDevice, BaseDeviceFactory as __BaseDeviceFactory


class RazerKeyboard(__RazerDevice):
    def __init__(self, serial, vid_pid=None, daemon_dbus=None):
        super(RazerKeyboard, self).__init__(serial, vid_pid=vid_pid, daemon_dbus=daemon_dbus)

        # Capabilities
        self._capabilities['game_mode_led'] = self._has_feature('razer.device.led.gamemode')
        self._capabilities['macro_mode_led'] = self._has_feature('razer.device.led.macromode', 'setMacroMode')
        self._capabilities['macro_mode_led_effect'] = self._has_feature('razer.device.led.macromode', 'setMacroEffect')
        self._capabilities['macro_tartarus_mode_modifier'] = self._has_feature('razer.device.macro', 'setModeModifier')

        # Setup base stuff if need be
        if self.has('game_mode_led'):
            self._dbus_interfaces['game_mode_led'] = _dbus.Interface(self._dbus, "razer.device.led.gamemode")

        if self.has('macro_mode_led'):
            self._dbus_interfaces['macro_mode_led'] = _dbus.Interface(self._dbus, "razer.device.led.macromode")


    @property
    def game_mode_led(self) -> bool:
        """
        Get game mode LED state

        :return: LED state
        :rtype: bool
        """
        if self.has('game_mode_led'):
            return self._dbus_interfaces['game_mode_led'].getGameMode()
        else:
            return False

    @game_mode_led.setter
    def game_mode_led(self, value:bool):
        """
        Set game mode LED state

        :param value: LED State
        :type value: bool
        """
        if self.has('game_mode_led'):
            if value:
                self._dbus_interfaces['game_mode_led'].setGameMode(True)
            else:
                self._dbus_interfaces['game_mode_led'].setGameMode(False)

    @property
    def macro_mode_led(self) -> bool:
        """
        Get macro mode LED state

        :return: LED state
        :rtype: bool
        """
        if self.has('macro_mode_led'):
            return self._dbus_interfaces['macro_mode_led'].getMacroMode()
        else:
            return False

    @macro_mode_led.setter
    def macro_mode_led(self, value:bool):
        """
        Set macro mode LED state

        :param value: LED State
        :type value: bool
        """
        if self.has('macro_mode_led'):
            if value:
                self._dbus_interfaces['macro_mode_led'].setMacroMode(True)
            else:
                self._dbus_interfaces['macro_mode_led'].setMacroMode(False)

    @property
    def macro_mode_led_effect(self) -> int:
        """
        Get macro mode LED effect

        Can get either blinking or static
        :return: Effect ID
        :rtype: int
        """
        if self.has('macro_mode_led_effect'):
            return self._dbus_interfaces['macro_mode_led'].getMacroEffect()
        else:
            return False

    @macro_mode_led_effect.setter
    def macro_mode_led_effect(self, value:int):
        """
        Set macro mode LED effect

        Can set to either MACRO_LED_STATIC or MACRO_LED_BLINK
        :param value: Effect ID
        :type value: int
        """
        if self.has('macro_mode_led_effect') and value in (MACRO_LED_STATIC, MACRO_LED_BLINK):
            self._dbus_interfaces['macro_mode_led'].setMacroEffect(value)


DEVICE_PID_MAP = {

}


class RazerKeyboardFactory(__BaseDeviceFactory):
    @staticmethod
    def get_device(serial, vid_pid=None, daemon_dbus=None):
        if vid_pid is None:
            pid = 0xFFFF
        else:
            pid = vid_pid[1]

        device_class = DEVICE_PID_MAP.get(pid, RazerKeyboard)
        return device_class(serial, vid_pid=vid_pid, daemon_dbus=daemon_dbus)



import time
import typing

from openrazer.client.devices.keyboard import RazerKeyboard
from openrazer.client.fx import Frame
from ore.core.effects import KeyboardEffect
from ore.core.utils import FpsLimiter, find_keyboard

MAX_FPS = 35


class KeyboardProgram(KeyboardEffect):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.fps = FpsLimiter(MAX_FPS)
        self.clear_every_frame = True

        self.keyboard = find_keyboard()

    def start(self):
        self.init(self.keyboard)
        self.loop(self.keyboard)

    def loop(self, kb: RazerKeyboard):

        stop = False
        last_timestamp = time.time()
        frame = 0

        while not stop:
            self.fps.start_frame()

            now = time.time()
            time_delta = now - last_timestamp
            last_timestamp = now

            if self.clear_every_frame:
                kb.fx.advanced.matrix.reset()

            stop = self.draw(kb, kb.fx.advanced.matrix, frame, time_delta)
            kb.fx.advanced.draw()

            if stop:
                break

            self.fps.wait_for_next_frame()
            frame += 1

    def init(self, kb: RazerKeyboard):
        raise NotImplementedError()

    def draw(self, kb: RazerKeyboard, matrix: Frame, frame_num: int, dt: float) -> typing.Any:
        raise NotImplementedError()



from openrazer.client.devices import RazerDevice as __RazerDevice


class RazerFirefly(__RazerDevice):
    def trigger_reactive(self) -> bool:
        """
        Trigger a reactive flash
        """
        return self._dbus_interfaces['device'].triggerReactive()



"""
Class to manage syncing of effects
"""
import inspect
import logging

class EffectSync(object):
    """
    Class which deals with receiving effect events from other devices
    """
    def __init__(self, parent, device_number):
        self._logger = logging.getLogger('razer.device{0}.effect_sync'.format(device_number))
        self._parent = parent
        self._parent.register_observer(self)

    def __del__(self):
        self.close()

    def close(self):
        self._parent.remove_observer(self)

    def notify(self, msg):
        """
        Receive notificatons from the device (we only care about effects)

        :param msg: Notification
        :type msg: tuple
        """
        if not isinstance(msg, tuple):
            self._logger.warning("Got msg that was not a tuple")
        elif msg[0] == 'effect':
            # We have a message directed at us
            # MSG format
            #  0         1       2             3
            # ('effect', Device, 'effectName', 'effectparams'...)
            # Device is the device the msg originated from (could be parent device)
            if msg[1] is not self._parent:
                # Msg from another device
                self.run_effect(msg[2], *msg[3:])

    def run_effect(self, effect_name, *args):
        """
        Run the specified effect with the given arguments

        :param effect_name: Name of the effect
        :type effect_name: str

        :param args: Arguments for the specified effect
        :type args: list
        """
        # Disable notifications
        self._parent.disable_notify = True

        try:
            # Does parent have method
            effect_func = getattr(self._parent, effect_name, None)
            if effect_func is not None:
                # We have method, does it have the correct num arguments
                actual_args = self.get_num_arguments(effect_func)
                if actual_args == len(args):
                    # method should be same
                    effect_func(*args)
                else:
                    # Method same but wrong args, try alternatives
                    if effect_name == 'setStatic':
                        # Could be static from chroma to non chroma
                        if actual_args == 0:
                            # Chroma -> BW
                            effect_func()
                        else:
                            # BW -> Chroma
                            effect_func(0x0, 0xFF, 0x00) # Green

            else:
                # We dont have method
                if effect_name == 'setPulsate':
                    # BW -> Chroma?
                    effect_func = getattr(self._parent, 'setBreathSingle', None)
                    if effect_func is not None:
                        effect_func(0x00, 0xFF, 0x00) # Green
                if effect_name in ('setBreathSingle', 'setBreathRandom', 'setBreathDual'):
                    # Chroma -> BW?
                    effect_func = getattr(self._parent, 'setPulsate', None)
                    if effect_func is not None:
                        effect_func() # Pulsate

        except Exception as err:
            self._logger.exception("Caught exception trying to sync effects.", exc_info=err)

        # Reenable notifications
        self._parent.disable_notify = False

    @staticmethod
    def get_num_arguments(func):
        """
        Get number of arguments in a function

        :param func: Function
        :type func: callable

        :return: Number of arguments
        :rtype: int
        """
        func_sig = inspect.signature(func)
        return len(func_sig.parameters)



"""
Headsets class
"""
import re

from openrazer_daemon.hardware.device_base import RazerDevice as __RazerDevice
from openrazer_daemon.dbus_services.dbus_methods import kraken as _dbus_kraken, chroma_keyboard as _dbus_chroma


class RazerKrakenClassic(__RazerDevice):
    """
    Class for the Razer Kraken 7.1 Chroma
    """
    EVENT_FILE_REGEX = re.compile(r'.*Razer_Kraken_7\.1_000000000000-event-if03')

    USB_VID = 0x1532
    USB_PID = 0x0501
    DEDICATED_MACRO_KEYS = False
    METHODS = ['get_firmware', 'get_device_name', 'get_device_type_headset', 'has_matrix', 'get_matrix_dims',
               'set_static_effect', 'set_none_effect', 'get_current_effect_kraken']

    RAZER_URLS = {
        "top_img": "https://assets.razerzone.com/eeimages/products/17519/02.png",
        "side_img": "https://assets.razerzone.com/eeimages/products/17519/03.png",
        "perspective_img": "https://assets.razerzone.com/eeimages/products/17519/01.png"
    }

    def __init__(self, *args, **kwargs):
        super(RazerKrakenClassic, self).__init__(*args, **kwargs)

    def _close(self):
        super(RazerKrakenClassic, self)._close()

    @staticmethod
    def decode_bitfield(bitfield):
        return {
            'state': (bitfield & 0x01) == 0x01,
            'breathing1': (bitfield & 0x02) == 0x02,
            'spectrum': (bitfield & 0x04) == 0x04,
            'sync': (bitfield & 0x08) == 0x08,
            'breathing2': (bitfield & 0x10) == 0x10,
            'breathing3': (bitfield & 0x20) == 0x20,
        }

    def _suspend_device(self):
        """
        Suspend the device

        Get the current brightness level, store it for later and then set the brightness to 0
        """
        self.suspend_args.clear()

        current_effect = _dbus_kraken.get_current_effect_kraken(self)
        dec = self.decode_bitfield(current_effect)

        if dec['state'] == 0x00:
            self.suspend_args['effect'] = 'none'
        elif dec['state'] == 0x01:
            self.suspend_args['effect'] = 'static'

        self.disable_notify = True
        _dbus_chroma.set_none_effect(self)
        self.disable_notify = False

    def _resume_device(self):
        """
        Resume the device

        Get the last known brightness and then set the brightness
        """

        effect = self.suspend_args.get('effect', '')

        self.disable_notify = True
        if effect == 'static':  # Static on classic is only 1 colour
            _dbus_chroma.set_static_effect(self, 0x00, 0x00, 0x00)

        self.disable_notify = False


class RazerKraken(__RazerDevice):
    """
    Class for the Razer Kraken 7.1 Chroma
    """
    EVENT_FILE_REGEX = re.compile(r'.*Razer_Kraken_7\.1_Chroma-event-if03')

    USB_VID = 0x1532
    USB_PID = 0x0504
    DEDICATED_MACRO_KEYS = False
    METHODS = ['get_firmware', 'get_device_name', 'get_device_type_headset', 'has_matrix', 'get_matrix_dims',
               'set_static_effect', 'set_spectrum_effect', 'set_none_effect', 'set_breath_single_effect',
               'get_current_effect_kraken', 'get_static_effect_args_kraken', 'get_breath_effect_args_kraken', 'set_custom_kraken']

    RAZER_URLS = {
        "top_img": "https://assets.razerzone.com/eeimages/products/17519/02.png",
        "side_img": "https://assets.razerzone.com/eeimages/products/17519/03.png",
        "perspective_img": "https://assets.razerzone.com/eeimages/products/17519/01.png"
    }

    def __init__(self, *args, **kwargs):
        super(RazerKraken, self).__init__(*args, **kwargs)

    def _close(self):
        super(RazerKraken, self)._close()

    @staticmethod
    def decode_bitfield(bitfield):
        return {
            'state': (bitfield & 0x01) == 0x01,
            'breathing1': (bitfield & 0x02) == 0x02,
            'spectrum': (bitfield & 0x04) == 0x04,
            'sync': (bitfield & 0x08) == 0x08,
            'breathing2': (bitfield & 0x10) == 0x10,
            'breathing3': (bitfield & 0x20) == 0x20,
        }

    def _suspend_device(self):
        """
        Suspend the device

        Get the current brightness level, store it for later and then set the brightness to 0
        """
        self.suspend_args.clear()

        current_effect = _dbus_kraken.get_current_effect_kraken(self)
        dec = self.decode_bitfield(current_effect)

        if dec['breathing1']:
            self.suspend_args['effect'] = 'breathing1'
            self.suspend_args['args'] = _dbus_kraken.get_breath_effect_args_kraken(self)
        elif dec['spectrum']:
            self.suspend_args['effect'] = 'spectrum'
        elif dec['state']:
            self.suspend_args['effect'] = 'static'
            self.suspend_args['args'] = _dbus_kraken.get_static_effect_args_kraken(self)

        self.disable_notify = True
        _dbus_chroma.set_none_effect(self)
        self.disable_notify = False

    def _resume_device(self):
        """
        Resume the device

        Get the last known brightness and then set the brightness
        """

        effect = self.suspend_args.get('effect', '')
        args = self.suspend_args.get('args', [])

        self.disable_notify = True
        if effect == 'spectrum':
            _dbus_chroma.set_spectrum_effect(self)
        elif effect == 'static':
            _dbus_chroma.set_static_effect(self, *args)
        elif effect == 'breathing1':
            _dbus_chroma.set_breath_single_effect(self, *args)

        self.disable_notify = False


class RazerKrakenV2(__RazerDevice):
    """
    Class for the Razer Kraken 7.1 V2
    """
    EVENT_FILE_REGEX = re.compile(r'.*Razer_Kraken_7\.1_V2_0+-event-if03')

    USB_VID = 0x1532
    USB_PID = 0x0510
    DEDICATED_MACRO_KEYS = False
    METHODS = ['get_firmware', 'get_device_name', 'get_device_type_headset', 'has_matrix', 'get_matrix_dims',
               'set_static_effect', 'set_spectrum_effect', 'set_none_effect', 'set_breath_single_effect', 'set_breath_dual_effect', 'set_breath_triple_effect',
               'get_current_effect_kraken', 'get_static_effect_args_kraken', 'get_breath_effect_args_kraken', 'set_custom_kraken']

    RAZER_URLS = {
        "top_img": "https://assets.razerzone.com/eeimages/products/26005/kraken71v2_gallery04-v2.png",
        "side_img": "https://assets.razerzone.com/eeimages/products/26005/kraken71v2_gallery01-v2.png",
        "perspective_img": "https://assets.razerzone.com/eeimages/products/26005/kraken71v2_gallery03-v2.png"
    }

    def __init__(self, *args, **kwargs):
        super(RazerKrakenV2, self).__init__(*args, **kwargs)

    def _close(self):
        super(RazerKrakenV2, self)._close()

    @staticmethod
    def decode_bitfield(bitfield):
        return {
            'state': (bitfield & 0x01) == 0x01,
            'breathing1': (bitfield & 0x02) == 0x02,
            'spectrum': (bitfield & 0x04) == 0x04,
            'sync': (bitfield & 0x08) == 0x08,
            'breathing2': (bitfield & 0x10) == 0x10,
            'breathing3': (bitfield & 0x20) == 0x20,
        }

    def _suspend_device(self):
        """
        Suspend the device

        Get the current brightness level, store it for later and then set the brightness to 0
        """
        self.suspend_args.clear()

        current_effect = _dbus_kraken.get_current_effect_kraken(self)
        dec = self.decode_bitfield(current_effect)

        if dec['breathing1']:
            self.suspend_args['effect'] = 'breathing1'
            self.suspend_args['args'] = _dbus_kraken.get_breath_effect_args_kraken(self)
        elif dec['breathing2']:
            self.suspend_args['effect'] = 'breathing2'
            self.suspend_args['args'] = _dbus_kraken.get_breath_effect_args_kraken(self)
        elif dec['breathing3']:
            self.suspend_args['effect'] = 'breathing3'
            self.suspend_args['args'] = _dbus_kraken.get_breath_effect_args_kraken(self)
        elif dec['spectrum']:
            self.suspend_args['effect'] = 'spectrum'
        elif dec['state']:
            self.suspend_args['effect'] = 'static'
            self.suspend_args['args'] = _dbus_kraken.get_static_effect_args_kraken(self)

        self.disable_notify = True
        _dbus_chroma.set_none_effect(self)
        self.disable_notify = False

    def _resume_device(self):
        """
        Resume the device

        Get the last known brightness and then set the brightness
        """

        effect = self.suspend_args.get('effect', '')
        args = self.suspend_args.get('args', [])

        self.disable_notify = True
        if effect == 'spectrum':
            _dbus_chroma.set_spectrum_effect(self)
        elif effect == 'static':
            _dbus_chroma.set_static_effect(self, *args)
        elif effect == 'breathing1':
            _dbus_chroma.set_breath_single_effect(self, *args)
        elif effect == 'breathing2':
            _dbus_chroma.set_breath_dual_effect(self, *args)
        elif effect == 'breathing3':
            _dbus_chroma.set_breath_triple_effect(self, *args)

        self.disable_notify = False



"""
BlackWidow Chroma Effects
"""
import math
import struct
from openrazer_daemon.dbus_services import endpoint


@endpoint('razer.device.power', 'getBattery', out_sig='d')
def get_battery(self):
    """
    Get mouse's battery level
    """
    self.logger.debug("DBus call get_battery")

    driver_path = self.get_driver_path('charge_level')

    with open(driver_path, 'r') as driver_file:
        battery_255 = float(driver_file.read().strip())
        if battery_255 < 0:
            return -1.0

        battery_100 = (battery_255 / 255) * 100
        return battery_100


@endpoint('razer.device.power', 'isCharging', out_sig='b')
def is_charging(self):
    """
    Get charging status
    """
    self.logger.debug("DBus call is_charging")

    driver_path = self.get_driver_path('charge_status')

    with open(driver_path, 'r') as driver_file:
        return bool(int(driver_file.read().strip()))


@endpoint('razer.device.power', 'setIdleTime', in_sig='q')
def set_idle_time(self, idle_time):
    """
    Set the idle time of the mouse in seconds

    :param idle_time: Idle time in seconds (unsigned short)
    :type idle_time: int
    """
    self.logger.debug("DBus call set_idle_time")

    driver_path = self.get_driver_path('device_idle_time')

    with open(driver_path, 'w') as driver_file:
        driver_file.write(str(idle_time))


@endpoint('razer.device.power', 'setLowBatteryThreshold', in_sig='y')
def set_low_battery_threshold(self, threshold):
    """
    Set the low battery threshold as a percentage

    :param threshold: Battery threshold as a percentage
    :type threshold: int
    """
    self.logger.debug("DBus call set_low_battery_threshold")

    driver_path = self.get_driver_path('charge_low_threshold')

    threshold = math.floor((threshold/100) * 255)

    with open(driver_path, 'w') as driver_file:
        driver_file.write(str(threshold))


@endpoint('razer.device.lighting.power', 'setChargeEffect', in_sig='y')
def set_charge_effect(self, charge_effect):
    """
    Set the charging effect.

    If 0x00 then it will use the current mouse's effect
    If 0x01 it will use the charge colour

    :param charge_effect: Charge effect
    :type charge_effect: int
    :return:
    """
    self.logger.debug("DBus call set_charge_effect")

    driver_path = self.get_driver_path('charge_effect')

    with open(driver_path, 'wb') as driver_file:
        driver_file.write(bytes([charge_effect]))


@endpoint('razer.device.lighting.power', 'setChargeColour', in_sig='yyy')
def set_charge_colour(self, red, green, blue):
    """
    Set the charge colour

    :param red: Red component
    :type red: int

    :param green: Green component
    :type green: int

    :param blue: Blue component
    :type blue: int
    """
    self.logger.debug("DBus call set_charge_colour")

    driver_path = self.get_driver_path('charge_colour')

    payload = bytes([red, green, blue])

    with open(driver_path, 'wb') as driver_file:
        driver_file.write(payload)


@endpoint('razer.device.dpi', 'setDPI', in_sig='qq')
def set_dpi_xy(self, dpi_x, dpi_y):
    """
    Set the DPI on the mouse, Takes in 4 bytes big-endian

    :param dpi_x: X DPI
    :type dpi_x: int
    :param dpi_y: Y DPI
    :type dpi_x: int
    """
    self.logger.debug("DBus call set_dpi_both")

    driver_path = self.get_driver_path('dpi')

    dpi_bytes = struct.pack('>HH', dpi_x, dpi_y)

    with open(driver_path, 'wb') as driver_file:
        driver_file.write(dpi_bytes)


@endpoint('razer.device.dpi', 'getDPI', out_sig='ai')
def get_dpi_xy(self):
    """
    get the DPI on the mouse

    :return: List of X, Y DPI
    :rtype: list of int
    """
    self.logger.debug("DBus call get_dpi_both")

    driver_path = self.get_driver_path('dpi')

    with open(driver_path, 'r') as driver_file:
        result = driver_file.read()
        dpi_x, dpi_y = [int(dpi) for dpi in result.strip().split(':')]

    return [dpi_x, dpi_y]


@endpoint('razer.device.dpi', 'maxDPI', out_sig='i')
def max_dpi(self):
    self.logger.debug("DBus call get_dpi_both")

    if hasattr(self, 'DPI_MAX'):
        return self.DPI_MAX

    else:
        return 500



@endpoint('razer.device.misc', 'setPollRate', in_sig='q')
def set_poll_rate(self, rate):
    """
    Set the DPI on the mouse, Takes in 4 bytes big-endian

    :param rate: Poll rate
    :type rate: int
    """
    self.logger.debug("DBus call set_poll_rate")

    if rate in (1000, 500, 125):
        driver_path = self.get_driver_path('poll_rate')

        with open(driver_path, 'w') as driver_file:
            driver_file.write(str(rate))
    else:
        self.logger.error("Poll rate %d is invalid", rate)


@endpoint('razer.device.misc', 'getPollRate', out_sig='i')
def get_poll_rate(self):
    """
    Get the polling rate from the device

    :return: Poll rate
    :rtype: int
    """
    self.logger.debug("DBus call get_poll_rate")

    driver_path = self.get_driver_path('poll_rate')

    with open(driver_path, 'r') as driver_file:
        result = driver_file.read()
        result = int(result.strip())

    return result





"""
Daemon class

This class is the main core of the daemon, this serves a basic dbus module to control the main bit of the daemon
"""
__version__ = '2.0.0'

import configparser
import logging
import logging.handlers
import os
import subprocess
import sys
import signal
import time
import tempfile
import setproctitle
import dbus.mainloop.glib
import dbus.service
import gi
gi.require_version('Gdk', '3.0')
import gi.repository
from gi.repository import GObject, GLib
from pyudev import Context, Monitor, MonitorObserver
import grp
import getpass
import json

import openrazer_daemon.hardware
from openrazer_daemon.dbus_services.service import DBusService
from openrazer_daemon.device import DeviceCollection
from openrazer_daemon.misc.screensaver_monitor import ScreensaverMonitor

class RazerDaemon(DBusService):
    """
    Daemon class

    This class sets up the main run loop which serves DBus messages. The logger is initialised
    in this module as well as finding and initialising devices.

    Serves the following functions via DBus
    * getDevices - Returns a list of serial numbers
    * enableTurnOffOnScreensaver - Starts/Continues the run loop on the screensaver thread
    * disableTurnOffOnScreensaver - Pauses the run loop on the screensaver thread
    """

    BUS_NAME = 'org.razer'

    def __init__(self, verbose=False, log_dir=None, console_log=False, run_dir=None, config_file=None, test_dir=None):

        setproctitle.setproctitle('openrazer-daemon')

        # Expanding ~ as python doesnt do it by default, also creating dirs if needed
        try:
            if log_dir is not None:
                log_dir = os.path.expanduser(log_dir)
                os.makedirs(log_dir, exist_ok=True)
            if run_dir is not None:
                run_dir = os.path.expanduser(run_dir)
                os.makedirs(run_dir, exist_ok=True)
        except NotADirectoryError as e:
            print("Failed to create {}".format(e.filename), file=sys.stderr)
            sys.exit(1)

        if config_file is not None:
            config_file = os.path.expanduser(config_file)
            if not os.path.exists(config_file):
                print("Config file {} does not exist.".format(config_file), file=sys.stderr)
                sys.exit(1)

        self._test_dir = test_dir
        self._run_dir = run_dir
        self._config_file = config_file
        self._config = configparser.ConfigParser()
        self.read_config(config_file)

        # Logging
        log_level = logging.INFO
        if verbose or self._config.getboolean('General', 'verbose_logging'):
            log_level = logging.DEBUG
        self.logger = self._create_logger(log_dir, log_level, console_log)

        # Check for plugdev group
        if not self._check_plugdev_group():
            self.logger.critical("User is not a member of the plugdev group")
            sys.exit(1)

        # Setup DBus to use gobject main loop
        dbus.mainloop.glib.threads_init()
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        DBusService.__init__(self, self.BUS_NAME, '/org/razer')

        self._init_signals()
        self._main_loop = GObject.MainLoop()

        # Listen for input events from udev
        self._init_udev_monitor()

        # Load Classes
        self._device_classes = openrazer_daemon.hardware.get_device_classes()

        self.logger.info("Initialising Daemon (v%s). Pid: %d", __version__, os.getpid())
        self._init_screensaver_monitor()

        self._razer_devices = DeviceCollection()
        self._load_devices(first_run=True)

        # Add DBus methods
        methods = {
            # interface, method, callback, in-args, out-args
            ('razer.devices', 'getDevices', self.get_serial_list, None, 'as'),
            ('razer.devices', 'supportedDevices', self.supported_devices, None, 's'),
            ('razer.devices', 'enableTurnOffOnScreensaver', self.enable_turn_off_on_screensaver, 'b', None),
            ('razer.devices', 'getOffOnScreensaver', self.get_off_on_screensaver, None, 'b'),
            ('razer.devices', 'syncEffects', self.sync_effects, 'b', None),
            ('razer.devices', 'getSyncEffects', self.get_sync_effects, None, 'b'),
            ('razer.daemon', 'version', self.version, None, 's'),
            ('razer.daemon', 'stop', self.stop, None, None),
        }

        for m in methods:
            self.logger.debug("Adding {}.{} method to DBus".format(m[0], m[1]))
            self.add_dbus_method(m[0], m[1], m[2], in_signature=m[3], out_signature=m[4])

        # TODO remove
        self.sync_effects(self._config.getboolean('Startup', 'sync_effects_enabled'))
        # TODO ======

    @dbus.service.signal('razer.devices')
    def device_removed(self):
        self.logger.debug("Emitted Device Remove Signal")

    @dbus.service.signal('razer.devices')
    def device_added(self):
        self.logger.debug("Emitted Device Added Signal")

    def _create_logger(self, log_dir, log_level, want_console_log):
        """
        Initializes a logger and returns it.

        :param log_dir: If not None, specifies the directory to create the
        log file in
        :param log_level: The log level of messages to print
        :param want_console_log: True if we should print to the console

        :rtype:logging.Logger
        """
        logger = logging.getLogger('razer')
        logger.setLevel(log_level)
        formatter = logging.Formatter('%(asctime)s | %(name)-30s | %(levelname)-8s | %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        # Dont propagate to default logger
        logger.propagate = 0

        if want_console_log:
            console_logger = logging.StreamHandler()
            console_logger.setLevel(log_level)
            console_logger.setFormatter(formatter)
            logger.addHandler(console_logger)

        if log_dir is not None:
            log_file = os.path.join(log_dir, 'razer.log')
            file_logger = logging.handlers.RotatingFileHandler(log_file, maxBytes=16777216, backupCount=10) # 16MiB
            file_logger.setLevel(log_level)
            file_logger.setFormatter(formatter)
            logger.addHandler(file_logger)

        return logger

    def _check_plugdev_group(self):
        """
        Check if the user is a member of the plugdev group. For the root
        user, this always returns True

        :rtype: bool
        """
        user = getpass.getuser()
        if user == 'root':
            return True

        try:
            plugdev_group = grp.getgrnam('plugdev')

            if user in plugdev_group.gr_mem:
                return True
        except KeyError:
            pass

        return False

    def _init_udev_monitor(self):
        self._udev_context = Context()
        udev_monitor = Monitor.from_netlink(self._udev_context)
        udev_monitor.filter_by(subsystem='hid')
        self._udev_observer = MonitorObserver(udev_monitor, callback=self._udev_input_event, name='device-monitor')

    def _init_screensaver_monitor(self):
        try:
            self._screensaver_monitor = ScreensaverMonitor(self)
            self._screensaver_monitor.monitoring = self._config.getboolean('Startup', 'devices_off_on_screensaver')
        except dbus.exceptions.DBusException as e:
            self.logger.error("Failed to init ScreensaverMonitor: {}".format(e))

    def _init_signals(self):
        """
        Heinous hack to properly handle signals on the mainloop. Necessary
        if we want to use the mainloop run() functionality.
        """
        def signal_action(signum):
            """
            Action to take when a signal is trapped
            """
            self.quit(signum)

        def idle_handler():
            """
            GLib idle handler to propagate signals
            """
            GLib.idle_add(signal_action, priority=GLib.PRIORITY_HIGH)

        def handler(*args):
            """
            Unix signal handler
            """
            signal_action(args[0])

        def install_glib_handler(sig):
            """
            Choose a compatible method and install the handler
            """
            unix_signal_add = None

            if hasattr(GLib, "unix_signal_add"):
                unix_signal_add = GLib.unix_signal_add
            elif hasattr(GLib, "unix_signal_add_full"):
                unix_signal_add = GLib.unix_signal_add_full

            if unix_signal_add:
                unix_signal_add(GLib.PRIORITY_HIGH, sig, handler, sig)
            else:
                print("Can't install GLib signal handler!")

        for sig in signal.SIGINT, signal.SIGTERM, signal.SIGHUP:
            signal.signal(sig, idle_handler)
            GLib.idle_add(install_glib_handler, sig, priority=GLib.PRIORITY_HIGH)

    def read_config(self, config_file):
        """
        Read in the config file and set the defaults

        :param config_file: Config file
        :type config_file: str or None
        """
        # Generate sections as trying to access a value even if a default exists will die if the section does not
        for section in ('General', 'Startup', 'Statistics'):
            self._config[section] = {}

        self._config['DEFAULT'] = {
            'verbose_logging': True,
            'sync_effects_enabled': True,
            'devices_off_on_screensaver': True,
            'key_statistics': False,
        }

        if config_file is not None and os.path.exists(config_file):
            self._config.read(config_file)

    def get_off_on_screensaver(self):
        """
        Returns if turn off on screensaver

        :return: Result
        :rtype: bool
        """
        return self._screensaver_monitor.monitoring

    def enable_turn_off_on_screensaver(self, enable):
        """
        Enable the turning off of devices when the screensaver is active
        """
        self._screensaver_monitor.monitoring = enable

    def supported_devices(self):
        result = {cls.__name__: (cls.USB_VID, cls.USB_PID) for cls in self._device_classes}

        return json.dumps(result)

    def version(self):
        """
        Get the daemon version

        :return: Version string
        :rtype: str
        """
        return __version__

    def suspend_devices(self):
        """
        Suspend all devices
        """
        for device in self._razer_devices:
            device.dbus.suspend_device()

    def resume_devices(self):
        """
        Resume all devices
        """
        for device in self._razer_devices:
            device.dbus.resume_device()

    def get_serial_list(self):
        """
        Get list of devices serials
        """
        serial_list = self._razer_devices.serials()
        self.logger.debug('DBus called get_serial_list')
        return serial_list

    def sync_effects(self, enabled):
        """
        Sync the effects across the devices

        :param enabled: True to sync effects
        :type enabled: bool
        """
        # Todo perhaps move logic to device collection
        for device in self._razer_devices.devices:
            device.dbus.effect_sync = enabled

    def get_sync_effects(self):
        """
        Sync the effects across the devices

        :return: True if any devices sync effects
        :rtype: bool
        """
        result = False

        for device in self._razer_devices.devices:
            result |= device.dbus.effect_sync

        return result

    def _load_devices(self, first_run=False):
        """
        Go through supported devices and load them

        Loops through the available hardware classes, loops through
        each device in the system and adds it if needs be.
        """
        if first_run:
            # Just some pretty output
            max_name_len = max([len(cls.__name__) for cls in self._device_classes]) + 2
            for cls in self._device_classes:
                format_str = 'Loaded device specification: {0:-<' + str(max_name_len) + '} ({1:04x}:{2:04X})'

                self.logger.debug(format_str.format(cls.__name__ + ' ', cls.USB_VID, cls.USB_PID))

        if self._test_dir is not None:
            device_list = os.listdir(self._test_dir)
            test_mode = True
        else:
            device_list = list(self._udev_context.list_devices(subsystem='hid'))
            test_mode = False

        device_number = 0
        for device in device_list:

            for device_class in self._device_classes:
                # Interoperability between generic list of 0000:0000:0000.0000 and pyudev
                if test_mode:
                    sys_name = device
                    sys_path = os.path.join(self._test_dir, device)
                else:
                    sys_name = device.sys_name
                    sys_path = device.sys_path

                if sys_name in self._razer_devices:
                    continue

                if device_class.match(sys_name, sys_path):  # Check it matches sys/ ID format and has device_type file
                    self.logger.info('Found device.%d: %s', device_number, sys_name)

                    # TODO add testdir support
                    # Basically find the other usb interfaces
                    device_match = sys_name.split('.')[0]
                    additional_interfaces = []
                    if not test_mode:
                        for alt_device in device_list:
                            if device_match in alt_device.sys_name and alt_device.sys_name != sys_name:
                                additional_interfaces.append(alt_device.sys_path)

                    # Checking permissions
                    test_file = os.path.join(sys_path, 'device_type')
                    file_group_id = os.stat(test_file).st_gid
                    file_group_name = grp.getgrgid(file_group_id)[0]

                    if os.getgid() != file_group_id and file_group_name != 'plugdev':
                        self.logger.critical("Could not access {0}/device_type, file is not owned by plugdev".format(sys_path))
                        break

                    razer_device = device_class(sys_path, device_number, self._config, testing=self._test_dir is not None, additional_interfaces=sorted(additional_interfaces))

                    # Wireless devices sometimes dont listen
                    count = 0
                    while count < 3:
                        # Loop to get serial, exit early if it gets one
                        device_serial = razer_device.get_serial()
                        if len(device_serial) > 0:
                            break
                        time.sleep(0.1)
                        count += 1
                    else:
                        logging.warning("Could not get serial for device {0}. Skipping".format(sys_name))
                        continue

                    self._razer_devices.add(sys_name, device_serial, razer_device)

                    device_number += 1

    def _add_device(self, device):
        """
        Add device event from udev

        :param device: Udev Device
        :type device: pyudev.device._device.Device
        """
        device_number = len(self._razer_devices)
        for device_class in self._device_classes:
            sys_name = device.sys_name
            sys_path = device.sys_path

            if sys_name in self._razer_devices:
                continue

            if device_class.match(sys_name, sys_path):  # Check it matches sys/ ID format and has device_type file
                self.logger.info('Found valid device.%d: %s', device_number, sys_name)
                razer_device = device_class(sys_path, device_number, self._config, testing=self._test_dir is not None)

                # Its a udev event so currently the device hasn't been chmodded yet
                time.sleep(0.2)

                # Wireless devices sometimes dont listen
                device_serial = razer_device.get_serial()

                if len(device_serial) > 0:
                    # Add Device
                    self._razer_devices.add(sys_name, device_serial, razer_device)
                    self.device_added()
                else:
                    logging.warning("Could not get serial for device {0}. Skipping".format(sys_name))

    def _remove_device(self, device):
        """
        Remove device event from udev

        :param device: Udev Device
        :type device: pyudev.device._device.Device
        """
        device_id = device.sys_name

        try:
            device = self._razer_devices[device_id]

            device.dbus.close()
            device.dbus.remove_from_connection()
            self.logger.warning("Removing %s", device_id)

            # Delete device
            del self._razer_devices[device.device_id]
            self.device_removed()

        except IndexError:  # Why didnt i set it up as KeyError
            # It will return "extra" events for the additional usb interfaces bound to the driver
            pass

    def _udev_input_event(self, device):
        """
        Function called by the Udev monitor (#observerPattern)

        :param device: Udev device
        :type device: pyudev.device._device.Device
        """
        self.logger.debug('Device event [%s]: %s', device.action, device.device_path)
        if device.action == 'add':
            self._add_device(device)
        elif device.action == 'remove':
            self._remove_device(device)

    def run(self):
        """
        Run the daemon
        """
        self.logger.info('Serving DBus')

        # Start listening for device changes
        self._udev_observer.start()

        # Start the mainloop
        try:
            self._main_loop.run()
        except KeyboardInterrupt:
            self.logger.debug('Shutting down')

    def stop(self):
        """
        Wrapper for quit
        """
        self.quit(None)

    def quit(self, signum):
        """
        Quit by stopping the main loop, observer, and screensaver thread
        """
        # pylint: disable=unused-argument
        if signum is None:
            self.logger.info('Stopping daemon.')
        else:
            self.logger.info('Stopping daemon on signal %d', signum)

        self._main_loop.quit()

        # Stop udev monitor
        self._udev_observer.send_stop()

        for device in self._razer_devices:
            device.dbus.close()
